import csv
import io
import os
import urllib.request

import numpy

from ..constants.image import C_HEIGHT
from ..constants.image import C_MD5_DIGEST
from ..constants.image import C_SCALING
from ..constants.image import C_WIDTH
from ..constants.measurement import C_CHANNEL, C_C, C_Z, C_T, C_OBJECTS_CHANNEL, C_OBJECTS_Z, C_OBJECTS_T
from ..constants.measurement import COLTYPE_FLOAT
from ..constants.measurement import COLTYPE_INTEGER
from ..constants.measurement import COLTYPE_VARCHAR
from ..constants.measurement import COLTYPE_VARCHAR_FILE_NAME
from ..constants.measurement import COLTYPE_VARCHAR_FORMAT
from ..constants.measurement import COLTYPE_VARCHAR_PATH_NAME
from ..constants.measurement import C_FILE_NAME
from ..constants.measurement import C_FRAME
from ..constants.measurement import C_METADATA
from ..constants.measurement import C_OBJECTS_FILE_NAME
from ..constants.measurement import C_OBJECTS_FRAME
from ..constants.measurement import C_OBJECTS_PATH_NAME
from ..constants.measurement import C_OBJECTS_SERIES
from ..constants.measurement import C_OBJECTS_URL
from ..constants.measurement import C_PATH_NAME
from ..constants.measurement import C_SERIES
from ..constants.measurement import C_URL
from ..constants.measurement import FTR_WELL
from ..constants.measurement import M_WELL
from ..constants.measurement import PATH_NAME_LENGTH
from ..constants.measurement import GROUP_LENGTH
from ..constants.module import IO_FOLDER_CHOICE_HELP_TEXT
from ..constants.modules.load_data import DIR_ALL
from ..constants.modules.load_data import IMAGE_CATEGORIES
from ..constants.modules.load_data import OBJECTS_CATEGORIES
from ..constants.modules.load_data import PATH_PADDING
from ..image import FileImage
from ..measurement import Measurements
from ..module import Module
from ..object import Objects
from ..preferences import NO_FOLDER_NAME
from ..preferences import URL_FOLDER_NAME
from ..preferences import get_data_file
from ..preferences import is_url_path
from ..setting import Binary
from ..setting import ValidationError
from ..setting.do_something import DoSomething
from ..setting.multichoice import MultiChoice
from ..setting.range import IntegerRange
from ..setting.text import Directory
from ..setting.text import Filename
from ..utilities.core.module.identify import add_object_count_measurements
from ..utilities.core.module.identify import add_object_location_measurements
from ..utilities.core.module.identify import get_object_measurement_columns
from ..utilities.core.modules.load_data import bad_sizes_warning
from ..utilities.core.modules.load_data import get_image_name
from ..utilities.core.modules.load_data import get_loaddata_type
from ..utilities.core.modules.load_data import get_objects_name
from ..utilities.core.modules.load_data import header_to_column
from ..utilities.core.modules.load_data import is_file_name_feature
from ..utilities.core.modules.load_data import is_objects_file_name_feature
from ..utilities.core.modules.load_data import is_objects_url_name_feature
from ..utilities.core.modules.load_data import is_url_name_feature
from ..utilities.image import convert_image_to_objects
from ..utilities.image import generate_presigned_url
from ..utilities.measurement import get_length_from_varchar
from ..utilities.measurement import is_well_column_token
from ..utilities.measurement import is_well_row_token
from ..utilities.pathname import pathname2url
from ..utilities.pathname import url2pathname

__doc__ = """\
LoadData
========

**LoadData** loads text or numerical data to be associated with images,
and can also load images specified by file names.

This module loads a file that supplies text or numerical data associated
with the images to be processed, e.g., sample names, plate names, well
identifiers, or even a list of image file names to be processed in the
analysis run. Please note that most researchers will prefer to use the
Input modules (i.e., **Images**,
**Metadata**, **NamesAndTypes** and **Groups**) to load images.

Note that 3D images to be analyzed volumetrically CANNOT be loaded
with LoadData; they must be loaded with the Input modules.

The module reads files in CSV (comma-separated values) format.
These files can be produced by saving a spreadsheet from Excel as
“Windows Comma Separated Values” file format. The lines of the file
represent the rows, and each field in a row is separated by a comma.
Text values may be optionally enclosed by double quotes. The
**LoadData** module uses the first row of the file as a header. The
fields in this row provide the labels for each column of data.
Subsequent rows provide the values for each image cycle.

There are many reasons why you might want to prepare a CSV file and load
it via **LoadData**. Below, we describe how the column nomenclature
allows for special functionality for some downstream modules:

-  *Columns whose name begins with Image\_FileName or Image\_PathName:*
   A column whose name begins with “Image\_FileName” or
   “Image\_PathName” can be used to supply the file name and path name
   (relative to the base folder) of an image that you want to load, which
   will override the settings in the Input modules (**Images**,
   **Metadata**, **NamesAndTypes** and **Groups**). For instance,
   “Image\_FileName\_CY3” would supply the file name for the CY3-stained
   image, and choosing the *Load images based on this data?* option
   allows the CY3 images to be selected later in the pipeline.
   “Image\_PathName\_CY3” would supply the path names for the
   CY3-stained images. The path name column is optional; if all image
   files are in the base folder, this column is not needed.

-  *Columns whose name begins with Image\_ObjectsFileName or
   Image\_ObjectsPathName:* The behavior of these columns is identical
   to that of “Image\_FileName” or “Image\_PathName” except that it is
   used to specify an image that you want to load as objects.

-  *Columns whose name begins with Metadata:* A column whose name begins
   with “Metadata” can be used to group or associate files loaded by
   **LoadData**.

   For instance, an experiment might require that images created on the
   same day use an illumination correction function calculated from all
   images from that day, and furthermore, that the date be captured in
   the file names for the individual image sets and in a CSV file
   specifying the illumination correction functions.

   In this case, if the illumination correction images are loaded with
   the **LoadData** module, the file should have a “Metadata\_Date”
   column which contains the date metadata tags. Similarly, if the
   individual images are loaded using the **LoadImages** module,
   **LoadImages** should be set to extract the metadata tag from the
   file names (see **LoadImages** for more details on how to do so). The
   pipeline will then match the individual image with their
   corresponding illumination correction functions based on matching
   “Metadata\_Date” tags. This is useful if the same data is associated
   with several images (for example, multiple images obtained from a
   single well).

-  *Columns whose name begins with Series or Frame:* A column whose
   name begins with “Series” or “Frame” refers to
   information about image stacks or movies. The name of the image
   within CellProfiler appears after an underscore character. For
   example, “Frame\_DNA” would supply the frame number for the
   movie/image stack file specified by the “Image\_FileName\_DNA” and
   “Image\_PathName\_DNA” columns.

   Using a CSV for loading frames and/or series from an movie/image
   stack allows you more flexibility in assembling image sets for
   operations that would difficult or impossible using the Input modules
   alone. For example, if you wanted to analyze a movie of 1,000 frames
   by computing the difference between frames, you could create two
   image columns in a CSV, one for loading frames 1,2,…,999, and the
   other for loading frames 2,3,…,1000. In this case, CellProfiler would
   load the frame and its predecessor for each cycle and **ImageMath**
   could be used to create the difference image for downstream use.

-  *Columns that contain dose-response or positive/negative control
   information:* The **CalculateStatistics** module can calculate
   metrics of assay quality for an experiment if provided with
   information about which images represent positive and negative
   controls and/or what dose of treatment has been used for which
   images. This information is provided to **CalculateStatistics** via
   the **LoadData** module, using particular formats described in the
   help for **CalculateStatistics**. Again, using **LoadData** is useful
   if the same data is associated with several images (for example,
   multiple images obtained from a single well).

-  *Columns with any other name:* Columns of data beginning with any
   other text will be loaded into CellProfiler and then
   exported as a per-image measurement along with
   CellProfiler-calculated data. This is a convenient way for you to add
   data from your own sources to the files exported by CellProfiler.

|

============ ============ ===============
Supports 2D? Supports 3D? Respects masks?
============ ============ ===============
YES          NO           YES
============ ============ ===============

See also
^^^^^^^^

See also the **Input** modules (i.e., **Images**, **Metadata**,
**NamesAndTypes** and **Groups**), **LoadImages** and
**CalculateStatistics**.

Example CSV file
^^^^^^^^^^^^^^^^

+-------------------------+-------------------------+-------------------+-----------------------+
| Image\_FileName\_FITC   | Image\_PathName\_FITC   | Metadata\_Plate   | Titration\_NaCl\_uM   |
+-------------------------+-------------------------+-------------------+-----------------------+
| “04923\_d1.tif”         | “2009-07-08”            | “P-12345”         | 750                   |
+-------------------------+-------------------------+-------------------+-----------------------+
| “51265\_d1.tif”         | “2009-07-09”            | “P-12345”         | 2750                  |
+-------------------------+-------------------------+-------------------+-----------------------+

After the first row of header information (the column names), the first
image-specific row specifies the file, “2009-07-08/04923\_d1.tif” for
the FITC image (2009-07-08 is the name of the subfolder that contains
the image, relative to the Default Input Folder). The plate metadata is
“P-12345” and the NaCl titration used in the well is 750 uM. The second
image-specific row has the values “2009-07-09/51265\_d1.tif”, “P-12345”
and 2750 uM. The NaCl titration for the image is available for modules
that use numeric metadata, such as **CalculateStatistics**; “Titration”
will be the category and “NaCl\_uM” will be the measurement.

Using metadata in LoadData
^^^^^^^^^^^^^^^^^^^^^^^^^^

If you would like to use the metadata-specific settings, please see
*Help > General help > Using metadata in CellProfiler* for more details
on metadata usage and syntax. Briefly, **LoadData** can use metadata
provided by the input CSV file for grouping similar images together for
the analysis run and for metadata-specfic options in other modules; see
the settings help for *Group images by metadata* and, if that setting is
selected, *Select metadata tags for grouping* for details.

Using MetaXpress-acquired images in CellProfiler
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

To produce a CSV file containing image location and metadata from a
`MetaXpress`_ imaging run, do the following:

-  Collect image locations from all files that match the string *.tif*
   in the desired image folder, one row per image.
-  Split up the image pathname and filename into separate data columns
   for **LoadData** to read.
-  Remove data rows corresponding to:

   -  Thumbnail images (do not contain imaging data)
   -  Duplicate images (will cause metadata mismatching)
   -  Corrupt files (will cause failure on image loading)

-  The image data table may be linked to metadata contained in plate
   maps. These plate maps should be stored as flat files, and may be
   updated periodically via queries to a laboratory information
   management system (LIMS) database.
-  The complete image location and metadata is written to a CSV file
   where the headers can easily be formatted to match **LoadData**\ ’s
   input requirements (see column descriptions above). Single plates
   split across multiple directories (which often occurs in MetaXpress)
   are written to separate files and then merged, thereby removing the
   discontinuity.

More tips on using LoadData
^^^^^^^^^^^^^^^^^^^^^^^^^^^

For a GUI-based approach to creating a proper CSV file for use with
**LoadData**, we suggest using `KNIME`_ or `Pipeline Pilot`_.

For more details on configuring CellProfiler (and LoadData in
particular) for a LIMS environment, please see our `wiki`_ on the
subject.

Measurements made by this module
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

-  *Pathname, Filename:* The full path and the filename of each image,
   if you requested image loading.
-  *Scaling:* The maximum possible intensity value for the image format.
-  *Height, Width:* The height and width of images loaded by this module.
-  Any additional per-image data loaded from the input file you provided.

.. _MetaXpress: http://www.moleculardevices.com/systems/high-content-imaging/metaxpress-high-content-image-acquisition-and-analysis-software
.. _KNIME: https://www.knime.com/about
.. _Pipeline Pilot: http://accelrys.com/products/pipeline-pilot/
.. _wiki: http://github.com/CellProfiler/CellProfiler/wiki/Adapting-CellProfiler-to-a-LIMS-environment
"""

"""Reserve extra space in pathnames for batch processing name rewrites"""

"""Cache of header columns for files"""
header_cache = {}


class LoadData(Module):
    module_name = "LoadData"
    category = "File Processing"
    variable_revision_number = 6

    def create_settings(self):
        self.csv_directory = Directory(
            "Input data file location",
            allow_metadata=False,
            support_urls=True,
            doc="""\
Select the folder containing the CSV file to be loaded. {IO_FOLDER_CHOICE_HELP_TEXT}
""".format(
                **{"IO_FOLDER_CHOICE_HELP_TEXT": IO_FOLDER_CHOICE_HELP_TEXT}
            ),
        )

        def get_directory_fn():
            """Get the directory for the CSV file name"""
            return self.csv_directory.get_absolute_path()

        def set_directory_fn(path):
            dir_choice, custom_path = self.csv_directory.get_parts_from_path(path)
            self.csv_directory.join_parts(dir_choice, custom_path)

        self.csv_file_name = Filename(
            "Name of the file",
            "None",
            doc="""Provide the file name of the CSV file containing the data you want to load.""",
            get_directory_fn=get_directory_fn,
            set_directory_fn=set_directory_fn,
            browse_msg="Choose CSV file",
            exts=[("Data file (*.csv)", "*.csv"), ("All files (*.*)", "*.*")],
        )

        self.browse_csv_button = DoSomething(
            "Press to view CSV file contents", "View...", self.browse_csv
        )

        self.wants_images = Binary(
            "Load images based on this data?",
            True,
            doc="""\
Select *{YES}* to have **LoadData** load images based on the
*Image\_FileName* column and the *Image\_PathName* column (if specified).
""".format(
                **{"YES": "Yes"}
            ),
        )

        self.rescale = Binary(
            "Rescale intensities?",
            True,
            doc="""\
This option determines whether image metadata should be used to rescale
the image’s intensities. Some image formats save the maximum possible
intensity value along with the pixel data. For instance, a microscope
might acquire images using a 12-bit A/D converter which outputs
intensity values between zero and 4095, but stores the values in a field
that can take values up to 65535.

Select *{YES}* to rescale the image intensity so that the camera's maximum
possible intensity value is rescaled to 1.0 (by dividing all pixels in
the image by the camera's maximum possible intensity value, as indicated by
image metadata, if available, else by the datatype of the image).

Select *{NO}* to ignore the image metadata and rescale the image to a
maximum of 1.0 by dividing by max possible intensity value of the
image datatype (e.g. 255 or 65535).
""".format(
                **{"YES": "Yes", "NO": "No"}
            ),
        )

        self.image_directory = Directory(
            "Base image location",
            dir_choices=DIR_ALL,
            allow_metadata=False,
            doc="""\
The parent (base) folder where images are located. If images are
contained in subfolders, then the file you load with this module should
contain a column with path names relative to the base image folder (see
the general help for this module for more details). You can choose among
the following options:

-  *Default Input Folder:* Use the Default Input Folder.
-  *Default Output Folder:* Use the Default Output Folder.
-  *None:* You have an *Image\_PathName* field that supplies an absolute
   path.
-  *Elsewhere…*: Use a particular folder you specify.""",
        )

        self.wants_image_groupings = Binary(
            "Group images by metadata?",
            False,
            doc="""\
Select *{YES}* to break the image sets in an experiment into groups.
Each set of files that share your selected metadata tags will be processed
together. For example, see **CreateBatchFiles** for details on submitting a
CellProfiler pipeline to a computing cluster for processing groups
separately, and see the **Groups** module for other examples.
""".format(
                **{"YES": "Yes"}
            ),
        )

        self.metadata_fields = MultiChoice(
            "Select metadata tags for grouping",
            None,
            doc="""\
*(Used only if images are to be grouped by metadata)*

Select the tags by which you want to group the image files. You can
select multiple tags. For example, if a set of images had metadata for
“Run”, “Plate”, “Well”, and “Site”, selecting *Run* and *Plate* will
create groups containing images that share the same [*Run*,\ *Plate*]
pair of tags.""",
        )

        self.wants_rows = Binary(
            "Process just a range of rows?",
            False,
            doc="""\
Select *{YES}* if you want to process a subset of the rows in the CSV
file. In the boxes below, enter the number of the row you want to begin processing
with in the box on the left. Then, enter the number of the row you want to
end processing with in the box on the right. Rows are numbered starting at 1
(but do not count the header line). **LoadData** will process up to and including the end row.
""".format(
                **{"YES": "Yes"}
            ),
        )

        self.row_range = IntegerRange(
            "Rows to process",
            (1, 100000),
            1,
            doc="""\
*(Used only if a range of rows is to be specified)*

Enter the row numbers of the first and last row to be processed.""",
        )

        def do_reload():
            global header_cache
            header_cache = {}
            try:
                self.open_csv()
            except:
                pass

        self.clear_cache_button = DoSomething(
            "Reload cached information",
            "Reload",
            do_reload,
            doc="""\
*(Used only if the CSV is loaded from a URL)*

Press this button to reload header information saved inside
CellProfiler. **LoadData** caches information about your .csv file in
its memory for efficiency. The information is reloaded if a modification
is detected. **LoadData** might fail to detect a modification on a file
accessed over the network and will fail to detect modifications on files
accessed through HTTP or FTP. In this case, you will have to use this
button to reload the header information after changing the file.

This button will never destroy any information on disk. It is always
safe to press it.""",
        )

    def settings(self):
        return [
            self.csv_directory,
            self.csv_file_name,
            self.wants_images,
            self.image_directory,
            self.wants_rows,
            self.row_range,
            self.wants_image_groupings,
            self.metadata_fields,
            self.rescale,
        ]

    def validate_module(self, pipeline):
        csv_path = self.csv_path

        if self.csv_directory.dir_choice != URL_FOLDER_NAME:
            if not os.path.isfile(csv_path):
                raise ValidationError(
                    "No such CSV file: %s" % csv_path, self.csv_file_name
                )

        try:
            self.open_csv()
        except IOError as e:
            import errno

            if e.errno == errno.EWOULDBLOCK:
                raise ValidationError(
                    "Another program (Excel?) is locking the CSV file %s."
                    % self.csv_path,
                    self.csv_file_name,
                )
            else:
                raise ValidationError(
                    "Could not open CSV file %s (error: %s)" % (self.csv_path, e),
                    self.csv_file_name,
                )

        try:
            self.get_header()
        except Exception as e:
            raise ValidationError(
                "The CSV file, %s, is not in the proper format."
                " See this module's help for details on CSV format. (error: %s)"
                % (self.csv_path, e),
                self.csv_file_name,
            )

    def validate_module_warnings(self, pipeline):
        """Check for potentially dangerous settings

        The best practice is to have a single LoadImages or LoadData module.
        """
        for module in pipeline.modules():
            if id(module) == id(self):
                return
            if isinstance(module, LoadData):
                raise ValidationError(
                    "Your pipeline has two or more LoadData modules.\n"
                    "The best practice is to have only one LoadData module.\n"
                    "Consider combining the CSV files from all of your\n"
                    "LoadData modules into one and using only a single\n"
                    "LoadData module",
                    self.csv_file_name,
                )

        # check that user has selected fields for grouping if grouping is turned on
        if self.wants_image_groupings.value and (
            len(self.metadata_fields.selections) == 0
        ):
            raise ValidationError(
                "Group images by metadata is True, but no metadata "
                "tags have been chosen for grouping.",
                self.metadata_fields,
            )

    def visible_settings(self):
        result = [self.csv_directory, self.csv_file_name, self.browse_csv_button]
        if self.csv_directory.dir_choice == URL_FOLDER_NAME:
            result += [self.clear_cache_button]
            self.csv_file_name.text = "URL of the file"
            self.csv_file_name.set_browsable(False)
        else:
            self.csv_file_name.text = "Name of the file"
            self.csv_file_name.set_browsable(True)
        result += [self.wants_images]
        if self.wants_images.value:
            result += [self.rescale, self.image_directory, self.wants_image_groupings]
            if self.wants_image_groupings.value:
                result += [self.metadata_fields]
                try:
                    fields = [
                        field[len("Metadata_") :]
                        for field in self.get_header()
                        if field.startswith("Metadata_")
                    ]
                    if self.has_synthetic_well_metadata():
                        fields += [FTR_WELL]
                    self.metadata_fields.choices = fields
                except:
                    self.metadata_fields.choices = ["No CSV file"]

        result += [self.wants_rows]
        if self.wants_rows.value:
            result += [self.row_range]
        return result

    @property
    def csv_path(self):
        """The path and file name of the CSV file to be loaded"""
        if get_data_file() is not None:
            return get_data_file()
        if self.csv_directory.dir_choice == URL_FOLDER_NAME:
            return self.csv_file_name.value

        path = self.csv_directory.get_absolute_path()
        return os.path.join(path, self.csv_file_name.value)

    @property
    def image_path(self):
        return self.image_directory.get_absolute_path()

    @property
    def legacy_field_key(self):
        """The key to use to retrieve the metadata from the image set list"""
        return "LoadTextMetadata_%d" % self.module_num

    def get_cache_info(self):
        """Get the cached information for the data file"""
        global header_cache
        entry = header_cache.get(self.csv_path, dict(ctime=0))
        if is_url_path(self.csv_path):
            if self.csv_path not in header_cache:
                header_cache[self.csv_path] = entry
            return entry
        ctime = os.stat(self.csv_path).st_ctime
        if ctime > entry["ctime"]:
            entry = header_cache[self.csv_path] = {}
            entry["ctime"] = ctime
        return entry

    def open_csv(self, do_not_cache=False):
        """Open the csv file or URL, returning a file descriptor"""
        global header_cache

        if is_url_path(self.csv_path):
            if self.csv_path not in header_cache:
                header_cache[self.csv_path] = {}
            entry = header_cache[self.csv_path]
            if "URLEXCEPTION" in entry:
                raise entry["URLEXCEPTION"]
            if "URLDATA" in entry:
                fd = io.StringIO(entry["URLDATA"])
            else:
                if do_not_cache:
                    raise RuntimeError("Need to fetch URL manually.")
                try:
                    url = generate_presigned_url(self.csv_path)
                    url_fd = urllib.request.urlopen(url)
                except Exception as e:
                    entry["URLEXCEPTION"] = e
                    raise e
                fd = io.StringIO()
                while True:
                    text = url_fd.read()
                    if isinstance(text, bytes):
                        text = text.decode()
                    if len(text) == 0:
                        break
                    fd.write(text)
                fd.seek(0)
                entry["URLDATA"] = fd.getvalue()
            return fd
        else:
            return open(self.csv_path, "rt")

    def browse_csv(self):
        import wx
        from cellprofiler.gui.utilities.icon import get_cp_icon

        try:
            fd = self.open_csv()
        except:
            wx.MessageBox("Could not read %s" % self.csv_path)
            return
        reader = csv.reader(fd)
        header = next(reader)
        frame = wx.Frame(wx.GetApp().frame, title=self.csv_path)
        sizer = wx.BoxSizer(wx.VERTICAL)
        frame.SetSizer(sizer)
        list_ctl = wx.ListCtrl(frame, style=wx.LC_REPORT)
        sizer.Add(list_ctl, 1, wx.EXPAND)
        for i, field in enumerate(header):
            list_ctl.InsertColumn(i, field)
        for line in reader:
            list_ctl.Append(
                [s if isinstance(s, str) else s for s in line[: len(header)]]
            )
        frame.SetMinSize((640, 480))
        frame.SetIcon(get_cp_icon())
        frame.Fit()
        frame.Show()

    def get_header(self, do_not_cache=False):
        """Read the header fields from the csv file

        Open the csv file indicated by the settings and read the fields
        of its first line. These should be the measurement columns.
        """
        entry = self.get_cache_info()
        if "header" in entry:
            return entry["header"]

        fd = self.open_csv(do_not_cache=do_not_cache)
        reader = csv.reader(fd)
        header = next(reader)
        fd.close()
        entry["header"] = [header_to_column(column) for column in header]
        return entry["header"]

    def get_image_names(self, do_not_cache=False):
        header = self.get_header(do_not_cache=do_not_cache)
        image_names = set(
            [
                get_image_name(field)
                for field in header
                if is_file_name_feature(field) or is_url_name_feature(field)
            ]
        )
        return list(image_names)

    def get_object_names(self, do_not_cache=False):
        header = self.get_header(do_not_cache=do_not_cache)
        object_names = set(
            [
                get_objects_name(field)
                for field in header
                if is_objects_file_name_feature(field)
                or is_objects_url_name_feature(field)
            ]
        )
        return list(object_names)

    def other_providers(self, group):
        """Get name providers from the CSV header"""
        if group == "imagegroup" and self.wants_images.value:
            try:
                # do not load URLs automatically
                return self.get_image_names(do_not_cache=True)
            except Exception as e:
                return []
        elif group == "objectgroup" and self.wants_images:
            try:
                # do not load URLs automatically
                return self.get_object_names(do_not_cache=True)
            except Exception as e:
                return []

        return []

    def is_image_from_file(self, image_name):
        """Return True if LoadData provides the given image name"""
        providers = self.other_providers("imagegroup")
        return image_name in providers

    def is_load_module(self):
        """LoadData can make image sets so it's a load module"""
        return True

    def prepare_run(self, workspace):
        pipeline = workspace.pipeline
        m = workspace.measurements
        assert isinstance(m, Measurements)
        """Load the CSV file at the outset and populate the image set list"""
        if pipeline.in_batch_mode():
            return True
        fd = self.open_csv()
        reader = csv.reader(fd)
        header = [header_to_column(column) for column in next(reader)]
        if self.wants_rows.value:
            # skip initial rows
            rows = []
            for idx, row in enumerate(reader):
                if idx + 1 < self.row_range.min:
                    continue
                if idx + 1 > self.row_range.max:
                    break
                if len(row) == 0:
                    continue
                row = [s if isinstance(s, str) else s for s in row]
                if len(row) != len(header):
                    raise ValueError(
                        "Row # %d has the wrong number of elements: %d. Expected %d"
                        % (idx, len(row), len(header))
                    )
                rows.append(row)
        else:
            rows = [
                [s if isinstance(s, str) else s for s in row]
                for row in reader
                if len(row) > 0
            ]
        fd.close()
        #
        # Check for correct # of columns
        #
        n_fields = len(header)
        for i, row in enumerate(rows):
            if len(row) < n_fields:
                text = (
                    "Error on line %d of %s.\n" '\n"%s"\n' "%d rows found, expected %d"
                ) % (i + 2, self.csv_file_name.value, ",".join(row), len(row), n_fields)
                raise ValueError(text)
            elif len(row) > n_fields:
                del row[n_fields:]
        #
        # Find the metadata, object_name and image_name columns
        #
        metadata_columns = {}
        object_columns = {}
        image_columns = {}
        well_row_column = well_column_column = well_well_column = None
        for i, column in enumerate(header):
            if column.find("_") == -1:
                category = ""
                feature = column
            else:
                category, feature = column.split("_", 1)
            if category in IMAGE_CATEGORIES:
                if feature not in image_columns:
                    image_columns[feature] = []
                image_columns[feature].append(i)
            elif category in OBJECTS_CATEGORIES:
                if feature not in object_columns:
                    object_columns[feature] = []
                object_columns[feature].append(i)
            else:
                metadata_columns[column] = i
                if category == C_METADATA:
                    if feature.lower() == FTR_WELL.lower():
                        well_well_column = i
                    elif is_well_row_token(feature):
                        well_row_column = i
                    elif is_well_column_token(feature):
                        well_column_column = i

        if (
            well_row_column is not None
            and well_column_column is not None
            and well_well_column is None
        ):
            # add a synthetic well column
            metadata_columns[M_WELL] = len(header)
            header.append(M_WELL)
            for row in rows:
                row.append(row[well_row_column] + row[well_column_column])
        if self.wants_images:
            #
            # Add synthetic object and image columns
            #
            if self.image_directory.dir_choice == NO_FOLDER_NAME:
                path_base = ""
            else:
                path_base = self.image_path
            for d, url_category, file_name_category, path_name_category in (
                (image_columns, C_URL, C_FILE_NAME, C_PATH_NAME,),
                (
                    object_columns,
                    C_OBJECTS_URL,
                    C_OBJECTS_FILE_NAME,
                    C_OBJECTS_PATH_NAME,
                ),
            ):
                for name in list(d.keys()):
                    url_column = file_name_column = path_name_column = None
                    for k in d[name]:
                        if header[k].startswith(url_category):
                            url_column = k
                        elif header[k].startswith(file_name_category):
                            file_name_column = k
                        elif header[k].startswith(path_name_category):
                            path_name_column = k
                    if url_column is None:
                        if file_name_column is None:
                            raise ValueError(
                                (
                                    "LoadData needs a %s_%s column to match the "
                                    "%s_%s column"
                                )
                                % (file_name_category, name, path_name_category, name)
                            )
                        #
                        # Add URL column
                        #
                        d[name].append(len(header))
                        url_feature = "_".join((url_category, name))
                        header.append(url_feature)
                        for row in rows:
                            if path_name_column is None:
                                fullname = os.path.join(
                                    path_base, row[file_name_column]
                                )
                            else:
                                row_path_name = os.path.join(
                                    path_base, row[path_name_column]
                                )
                                fullname = os.path.join(
                                    row_path_name, row[file_name_column]
                                )
                                row[path_name_column] = row_path_name
                            url = pathname2url(fullname)
                            row.append(url)
                        if path_name_column is None:
                            #
                            # Add path column
                            #
                            d[name].append(len(header))
                            path_feature = "_".join((path_name_category, name))
                            header.append(path_feature)
                            for row in rows:
                                row.append(path_base)
                    elif path_name_column is None and file_name_column is None:
                        #
                        # If the .csv just has URLs, break the URL into
                        # path and file names
                        #
                        path_feature = "_".join((path_name_category, name))
                        path_name_column = len(header)
                        header.append(path_feature)

                        file_name_feature = "_".join((file_name_category, name))
                        file_name_column = len(header)
                        header.append(file_name_feature)
                        for row in rows:
                            url = row[url_column]
                            idx = url.rfind("/")
                            if idx == -1:
                                idx = url.rfind(":")
                                if idx == -1:
                                    row += ["", url]
                                else:
                                    row += [url[: (idx + 1)], url[(idx + 1) :]]
                            else:
                                row += [url[:idx], url[(idx + 1) :]]

        column_type = {}
        for column in self.get_measurement_columns(pipeline):
            column_type[column[1]] = column[2]

        previous_column_types = dict(
            [
                (c[1], c[2])
                for c in pipeline.get_measurement_columns(self)
                if c[0] == "Image"
            ]
        )
        #
        # Arrange the metadata into columns
        #
        columns = {}
        for index, feature in enumerate(header):
            c = []
            columns[feature] = c
            for row in rows:
                value = row[index]
                if feature in column_type:
                    datatype = column_type[feature]
                else:
                    datatype = previous_column_types[feature]
                if datatype == COLTYPE_INTEGER:
                    value = int(value)
                elif datatype == COLTYPE_FLOAT:
                    value = float(value)
                c.append(value)

        if len(metadata_columns) > 0:
            # Reorder the rows by matching metadata against previous metadata
            # (for instance, to assign metadata values to images from
            #  loadimages)
            #
            image_numbers = m.match_metadata(
                list(metadata_columns.keys()),
                [columns[k] for k in list(metadata_columns.keys())],
            )
            image_numbers = numpy.array(image_numbers, int).flatten()
            max_image_number = numpy.max(image_numbers)
            new_columns = {}
            for key, values in list(columns.items()):
                new_values = [None] * max_image_number
                for image_number, value in zip(image_numbers, values):
                    new_values[image_number - 1] = value
                new_columns[key] = new_values
            columns = new_columns
        for feature, values in list(columns.items()):
            m.add_all_measurements("Image", feature, values)
        if self.wants_image_groupings and len(self.metadata_fields.selections) > 0:
            keys = ["_".join((C_METADATA, k)) for k in self.metadata_fields.selections]
            m.set_grouping_tags(keys)
            groupkeys, groupvals = self.get_groupings(workspace)
            group_lengths = []
            for eachval in groupvals:
                group_lengths += [len(eachval[1])] * len(eachval[1])
            m.add_all_measurements(
                "Image", GROUP_LENGTH, group_lengths,
            )
        else:
            group_lengths = [len(rows)] * len(rows)
            m.add_all_measurements(
                "Image", GROUP_LENGTH, group_lengths,
            )

        return True

    def prepare_to_create_batch(self, workspace, fn_alter_path):
        """Prepare to create a batch file

        This function is called when CellProfiler is about to create a
        file for batch processing. It will pickle the image set list's
        "legacy_fields" dictionary. This callback lets a module prepare for
        saving.

        pipeline - the pipeline to be saved
        image_set_list - the image set list to be saved
        fn_alter_path - this is a function that takes a pathname on the local
                        host and returns a pathname on the remote host. It
                        handles issues such as replacing backslashes and
                        mapping mountpoints. It should be called for every
                        pathname stored in the settings or legacy fields.
        """

        if self.wants_images:
            m = workspace.measurements
            assert isinstance(m, Measurements)
            image_numbers = m.get_image_numbers()
            all_image_features = m.get_feature_names("Image")
            for url_category, file_category, path_category, names in (
                (C_URL, C_FILE_NAME, C_PATH_NAME, self.get_image_names(),),
                (
                    C_OBJECTS_URL,
                    C_OBJECTS_FILE_NAME,
                    C_OBJECTS_PATH_NAME,
                    self.get_object_names(),
                ),
            ):
                for name in names:
                    url_feature = "_".join((url_category, name))
                    path_feature = "_".join((path_category, name))
                    if path_feature not in all_image_features:
                        path_feature = None
                    file_feature = "_".join((file_category, name))
                    if file_feature not in all_image_features:
                        file_feature = None
                    urls = m.get_measurement(
                        "Image", url_feature, image_set_number=image_numbers,
                    )
                    for image_number, url in zip(image_numbers, urls):
                        url = url
                        if url.lower().startswith("file:"):
                            fullname = url2pathname(url)
                            fullname = fn_alter_path(fullname)
                            path, filename = os.path.split(fullname)
                            url = str(pathname2url(fullname))
                            m.add_measurement(
                                "Image",
                                url_feature,
                                url,
                                image_set_number=image_number,
                            )
                            if file_feature is not None:
                                m.add_measurement(
                                    "Image",
                                    file_feature,
                                    filename,
                                    image_set_number=image_number,
                                )
                            if path_feature is not None:
                                m.add_measurement(
                                    "Image",
                                    path_feature,
                                    path,
                                    image_set_number=image_number,
                                )

        self.csv_directory.alter_for_create_batch_files(fn_alter_path)
        self.image_directory.alter_for_create_batch_files(fn_alter_path)
        return True

    def fetch_provider(self, name, measurements, is_image_name=True):
        if is_image_name:
            url_feature = f"{C_URL}_{name}"
            series_feature = f"{C_SERIES}_{name}"
            frame_feature = f"{C_FRAME}_{name}"
            channel_feature = f"{C_C}_{name}"
            plane_feature = f"{C_Z}_{name}"
            timepoint_feature = f"{C_T}_{name}"
        else:
            url_feature = f"{C_OBJECTS_URL}_{name}"
            series_feature = f"{C_OBJECTS_SERIES}_{name}"
            frame_feature = f"{C_OBJECTS_FRAME}_{name}"
            channel_feature = f"{C_OBJECTS_CHANNEL}_{name}"
            plane_feature = f"{C_OBJECTS_Z}_{name}"
            timepoint_feature = f"{C_OBJECTS_T}_{name}"
        url = measurements.get_measurement("Image", url_feature)
        full_filename = url2pathname(url)
        path, filename = os.path.split(full_filename)
        keyword_args = {}
        for feature in (series_feature, frame_feature, channel_feature,
                        plane_feature, timepoint_feature):
            if measurements.has_feature("Image", feature) or (
                is_image_name
                and feature == C_CHANNEL
                and measurements.has_feature("Image", feature)
            ):
                val = measurements["Image", feature]
                if numpy.isnan(val):
                    val = None
                keyword_args[feature] = val
            else:
                keyword_args[feature] = None
        return FileImage(
            name,
            path,
            filename,
            rescale_range=None,
            metadata_rescale=self.rescale.value and is_image_name,
            series=keyword_args[series_feature],
            index=keyword_args[frame_feature],
            channel=keyword_args[channel_feature],
            z=keyword_args[plane_feature],
            t=keyword_args[timepoint_feature],
        )

    def run(self, workspace):
        """Populate the images and objects"""
        m = workspace.measurements
        assert isinstance(m, Measurements)
        image_set = workspace.image_set
        object_set = workspace.object_set
        statistics = []
        features = [
            x[1]
            for x in self.get_measurement_columns(workspace.pipeline)
            if x[0] == "Image"
        ]

        if self.wants_images:
            #
            # Load the image. Calculate the MD5 hash of every image
            #
            image_size = None
            for image_name in self.other_providers("imagegroup"):
                provider = self.fetch_provider(image_name, m)
                image_set.add_provider(provider)
                image = image_set.get_image(image_name)
                pixel_data = image.pixel_data
                m.add_image_measurement(
                    "_".join((C_MD5_DIGEST, image_name)), provider.get_md5_hash(m),
                )
                m.add_image_measurement(
                    "_".join((C_SCALING, image_name)), image.scale,
                )
                m.add_image_measurement(
                    "_".join((C_HEIGHT, image_name)), int(pixel_data.shape[0]),
                )
                m.add_image_measurement(
                    "_".join((C_WIDTH, image_name)), int(pixel_data.shape[1]),
                )
                if image_size is None:
                    image_size = tuple(pixel_data.shape[:2])
                    first_filename = image.file_name
                elif tuple(pixel_data.shape[:2]) != image_size:
                    warning = bad_sizes_warning(
                        image_size, first_filename, pixel_data.shape, image.file_name
                    )
                    if self.show_window:
                        workspace.display_data.warning = warning
                    else:
                        print(warning)
                        #
                        # Process any object tags
                        #
            objects_names = self.get_object_names()
            for objects_name in objects_names:
                provider = self.fetch_provider(objects_name, m, is_image_name=False)
                image = provider.provide_image(workspace.image_set)
                pixel_data = convert_image_to_objects(image.pixel_data)
                o = Objects()
                o.segmented = pixel_data
                object_set.add_objects(o, objects_name)
                add_object_count_measurements(m, objects_name, o.count)
                add_object_location_measurements(m, objects_name, pixel_data)

        for feature_name in sorted(features):
            value = m.get_measurement("Image", feature_name)
            statistics.append((feature_name, value))

        if self.show_window:
            workspace.display_data.statistics = statistics

    def display(self, workspace, figure):
        # if hasattr(workspace.display_data, "warning"):
        #     from cellprofiler.gui.errordialog import show_warning
        #     show_warning("Images have different sizes",
        #                  workspace.display_data.warning,
        #                  cpprefs.get_show_report_bad_sizes_dlg,
        #                  cpprefs.set_show_report_bad_sizes_dlg)

        figure.set_subplots((1, 1))
        figure.subplot_table(0, 0, workspace.display_data.statistics)

    def get_groupings(self, workspace):
        """Return the image groupings of the image sets

        See CPModule for documentation
        """
        if (
            self.wants_images.value
            and self.wants_image_groupings.value
            and len(self.metadata_fields.selections) > 0
        ):
            keys = ["_".join((C_METADATA, k)) for k in self.metadata_fields.selections]
            if len(keys) == 0:
                return None
            m = workspace.measurements
            assert isinstance(m, Measurements)
            return keys, m.get_groupings(keys)
        return None

    def get_measurement_columns(self, pipeline):
        """Return column definitions for measurements produced by this module"""
        entry = None
        try:
            entry = self.get_cache_info()
            if "measurement_columns" in entry:
                return entry["measurement_columns"]
            fd = self.open_csv()
            reader = csv.reader(fd)
            header = [header_to_column(x) for x in next(reader)]
        except:
            if entry is not None:
                entry["measurement_columns"] = []
            return []
        previous_columns = pipeline.get_measurement_columns(self)
        previous_fields = set([x[1] for x in previous_columns if x[0] == "Image"])
        already_output = [x in previous_fields for x in header]
        coltypes = [COLTYPE_INTEGER] * len(header)
        #
        # Make sure the well_column column type is a string
        #
        for i in range(len(header)):
            if header[i].startswith(C_METADATA + "_") and is_well_column_token(
                header[i].split("_")[1]
            ):
                coltypes[i] = COLTYPE_VARCHAR
            if any(
                [
                    header[i].startswith(x)
                    for x in (
                        C_PATH_NAME,
                        C_FILE_NAME,
                        C_OBJECTS_FILE_NAME,
                        C_OBJECTS_PATH_NAME,
                        C_URL,
                        C_OBJECTS_URL,
                    )
                ]
            ):
                coltypes[i] = COLTYPE_VARCHAR

        collen = [0] * len(header)
        key_is_path_or_file_name = [
            (
                key.startswith(C_PATH_NAME)
                or key.startswith(C_FILE_NAME)
                or key.startswith(C_OBJECTS_FILE_NAME)
                or key.startswith(C_OBJECTS_PATH_NAME)
            )
            for key in header
        ]
        key_is_path_or_url = [
            (
                key.startswith(C_PATH_NAME)
                or key.startswith(C_OBJECTS_PATH_NAME)
                or key.startswith(C_URL)
                or key.startswith(C_OBJECTS_URL)
            )
            for key in header
        ]

        for row in reader:
            if len(row) > len(header):
                row = row[: len(header)]
            for index, field in enumerate(row):
                if already_output[index]:
                    continue
                if (not self.wants_images) and key_is_path_or_file_name[index]:
                    continue
                try:
                    len_field = len(field)
                except TypeError:
                    field = str(field)
                    len_field = len(field)
                if key_is_path_or_url[index]:
                    # Account for possible rewrite of the pathname
                    # in batch data
                    len_field = max(PATH_NAME_LENGTH, len_field + PATH_PADDING,)
                if coltypes[index] != COLTYPE_VARCHAR:
                    ldtype = get_loaddata_type(field)
                    if coltypes[index] == COLTYPE_INTEGER:
                        coltypes[index] = ldtype
                    elif coltypes[index] == COLTYPE_FLOAT and ldtype != COLTYPE_INTEGER:
                        coltypes[index] = ldtype

                if collen[index] < len(field):
                    collen[index] = len(field)

        for index in range(len(header)):
            if coltypes[index] == COLTYPE_VARCHAR:
                coltypes[index] = COLTYPE_VARCHAR_FORMAT % collen[index]

        image_names = self.other_providers("imagegroup")
        result = [
            ("Image", colname, coltype)
            for colname, coltype in zip(header, coltypes)
            if colname not in previous_fields
        ]
        if self.wants_images:
            for feature, coltype in (
                (C_URL, COLTYPE_VARCHAR_PATH_NAME,),
                (C_PATH_NAME, COLTYPE_VARCHAR_PATH_NAME,),
                (C_FILE_NAME, COLTYPE_VARCHAR_FILE_NAME,),
                (C_MD5_DIGEST, COLTYPE_VARCHAR_FORMAT % 32,),
                (C_SCALING, COLTYPE_FLOAT,),
                (C_HEIGHT, COLTYPE_INTEGER,),
                (C_WIDTH, COLTYPE_INTEGER,),
            ):
                for image_name in image_names:
                    measurement = feature + "_" + image_name
                    if not any([measurement == c[1] for c in result]):
                        result.append(("Image", measurement, coltype,))
            #
            # Add the object features
            #
            for object_name in self.get_object_names():
                result += get_object_measurement_columns(object_name)
                for feature, coltype in (
                    (C_OBJECTS_URL, COLTYPE_VARCHAR_PATH_NAME,),
                    (C_OBJECTS_PATH_NAME, COLTYPE_VARCHAR_PATH_NAME,),
                    (C_OBJECTS_FILE_NAME, COLTYPE_VARCHAR_FILE_NAME,),
                ):
                    mname = C_OBJECTS_URL + "_" + object_name
                    result.append(("Image", mname, coltype))
        #
        # Try to make a well column out of well row and well column
        #
        well_column = None
        well_row_column = None
        well_col_column = None
        for column in result:
            if not column[1].startswith(C_METADATA + "_"):
                continue
            category, feature = column[1].split("_", 1)
            if is_well_column_token(feature):
                well_col_column = column
            elif is_well_row_token(feature):
                well_row_column = column
            elif feature.lower() == FTR_WELL.lower():
                well_column = column
        if (
            well_column is None
            and well_row_column is not None
            and well_col_column is not None
        ):
            length = get_length_from_varchar(well_row_column[2])
            length += get_length_from_varchar(well_col_column[2])
            result += [
                (
                    "Image",
                    "_".join((C_METADATA, FTR_WELL,)),
                    COLTYPE_VARCHAR_FORMAT % length,
                )
            ]
        entry["measurement_columns"] = result
        return result

    def has_synthetic_well_metadata(self):
        """Determine if we should synthesize a well metadata feature

        """
        fields = self.get_header()
        has_well_col = False
        has_well_row = False
        for field in fields:
            if not field.startswith(C_METADATA + "_"):
                continue
            category, feature = field.split("_", 1)
            if is_well_column_token(feature):
                has_well_col = True
            elif is_well_row_token(feature):
                has_well_row = True
            elif feature.lower() == FTR_WELL.lower():
                return False
        return has_well_col and has_well_row

    def get_categories(self, pipeline, object_name):
        try:
            columns = self.get_measurement_columns(pipeline)
            result = set(
                [
                    column[1].split("_")[0]
                    for column in columns
                    if column[0] == object_name
                ]
            )
            return list(result)
        except:
            return []

    def get_measurements(self, pipeline, object_name, category):
        columns = self.get_measurement_columns(pipeline)
        return [
            feature
            for c, feature in [
                column[1].split("_", 1)
                for column in columns
                if column[0] == object_name and column[1].startswith(category + "_")
            ]
        ]

    def change_causes_prepare_run(self, setting):
        """Check to see if changing the given setting means you have to restart

        Some settings, esp in modules like LoadImages, affect more than
        the current image set when changed. For instance, if you change
        the name specification for files, you have to reload your image_set_list.
        Override this and return True if changing the given setting means
        that you'll have to do "prepare_run".
        """
        if self.wants_images or setting == self.wants_images:
            return True
        return False

    def upgrade_settings(self, setting_values, variable_revision_number, module_name):

        dir_default_image = "Default input folder"
        dir_default_output = "Default Output Folder"

        if variable_revision_number == 1:
            setting_values = setting_values + ["No", ""]
            variable_revision_number = 2

        if variable_revision_number == 2:
            if setting_values[0].startswith("Default Image"):
                setting_values = [dir_default_image] + setting_values[1:]
            elif setting_values[0].startswith("Default Output"):
                setting_values = [dir_default_output] + setting_values[1:]
            if setting_values[4].startswith("Default Image"):
                setting_values = (
                    setting_values[:4] + [dir_default_image] + setting_values[5:]
                )
            elif setting_values[4].startswith("Default Output"):
                setting_values = (
                    setting_values[:4] + [dir_default_output] + setting_values[5:]
                )
            variable_revision_number = 3
        if variable_revision_number == 3:
            module_name = self.module_name

        if variable_revision_number == 3:
            # directory choice, custom directory merged
            # input_directory_choice, custom_input_directory merged
            (
                csv_directory_choice,
                csv_custom_directory,
                csv_file_name,
                wants_images,
                image_directory_choice,
                image_custom_directory,
                wants_rows,
                row_range,
                wants_image_groupings,
                metadata_fields,
            ) = setting_values
            csv_directory = Directory.static_join_string(
                csv_directory_choice, csv_custom_directory
            )
            image_directory = Directory.static_join_string(
                image_directory_choice, image_custom_directory
            )
            setting_values = [
                csv_directory,
                csv_file_name,
                wants_images,
                image_directory,
                wants_rows,
                row_range,
                wants_image_groupings,
                metadata_fields,
            ]
            variable_revision_number = 4

        # Standardize input/output directory name references
        setting_values = list(setting_values)
        for index in (0, 3):
            setting_values[index] = Directory.upgrade_setting(setting_values[index])

        if variable_revision_number == 4:
            (
                csv_directory,
                csv_file_name,
                wants_images,
                image_directory,
                wants_rows,
                row_range,
                wants_image_groupings,
                metadata_fields,
            ) = setting_values
            (dir_choice, custom_dir,) = Directory.split_string(csv_directory)
            if dir_choice == URL_FOLDER_NAME:
                csv_file_name = custom_dir + "/" + csv_file_name
                csv_directory = Directory.static_join_string(dir_choice, "")
            setting_values = [
                csv_directory,
                csv_file_name,
                wants_images,
                image_directory,
                wants_rows,
                row_range,
                wants_image_groupings,
                metadata_fields,
            ]
            variable_revision_number = 5
        if variable_revision_number == 5:
            # Added rescaling option
            setting_values = setting_values + ["Yes"]
            variable_revision_number = 6
        return setting_values, variable_revision_number


LoadText = LoadData



# coding=utf-8
import itertools
import logging

from cellprofiler_core.constants.measurement import C_SERIES_NAME, C_C, C_SERIES, C_Z, C_T
from cellprofiler_core.pipeline import ImagePlane
from cellprofiler_core.constants.module import FILTER_RULES_BUTTONS_HELP
from cellprofiler_core.constants.modules.images import FILTER_CHOICE_ALL
from cellprofiler_core.constants.modules.images import FILTER_CHOICE_CUSTOM
from cellprofiler_core.constants.modules.images import FILTER_CHOICE_IMAGES
from cellprofiler_core.constants.modules.images import FILTER_CHOICE_NONE
from cellprofiler_core.constants.modules.images import FILTER_DEFAULT
from cellprofiler_core.module import Module
from cellprofiler_core.setting import FileCollectionDisplay
from cellprofiler_core.setting import Binary
from cellprofiler_core.setting import PathListDisplay
from cellprofiler_core.setting.choice import Choice
from cellprofiler_core.setting.do_something import PathListRefreshButton
from cellprofiler_core.setting.do_something import PathListExtractButton
from cellprofiler_core.setting.filter import DirectoryPredicate
from cellprofiler_core.setting.filter import ExtensionPredicate
from cellprofiler_core.setting.filter import FilePredicate
from cellprofiler_core.setting.filter import Filter
from cellprofiler_core.utilities.image import image_resource
from cellprofiler_core.utilities.image import is_image


LOGGER = logging.getLogger(__name__)

__doc__ = """\
Images
======

The **Images** module allows you to specify the location of files to be
analyzed by the pipeline; setting this module correctly is the first
step in creating a new project in CellProfiler. These files can be
located on your hard drive, on a networked computer elsewhere, or
accessible with a URL. You can also provide rules to specify only those
files that you want analyzed out of a larger collection (for example,
from a folder containing both images for analysis and non-image files
that should be disregarded).

|

============ ============ ===============
Supports 2D? Supports 3D? Respects masks?
============ ============ ===============
YES          YES          NO
============ ============ ===============

What is a “digital image”?
^^^^^^^^^^^^^^^^^^^^^^^^^^

A *digital image* is a set of numbers arranged into a two-dimensional
format of rows and columns; a pixel refers to the row/column location of
a particular point in the image. Pixels in grayscale or monochrome
(black/white) images contain a single intensity value, whereas in color
images, each pixel contains a red, green, and blue (RGB) triplet of
intensity values. Additionally, the term image can be used as short-hand
for an image sequence, that is, an image collection such as a time-lapse
series (2-D + *t*), confocal Z-stacks (3-D), etc.

CellProfiler can read a wide variety of image formats by using a library
called Bio-Formats; see `here`_ for the formats available. Some image
formats are better than others for use in image analysis. Some are
`“lossy”`_ (information is lost in the conversion to the format) like
most JPG/JPEG files; others are `“lossless”`_ (no image information is
lost). For image analysis purposes, a lossless format like TIF or PNG is
recommended.

What do I need as input?
^^^^^^^^^^^^^^^^^^^^^^^^

The most straightforward way to provide image files to the **Images**
module is to simply drag-and-drop them on the file list panel (the blank
space indicated by the text “Drop files and folders here”).

.. image:: {IMG_PANEL_BLANK}
   :width: 100%

Using the file explorer tool of your choice (e.g., Explorer in Windows,
Finder in Mac), you can drag-and-drop individual files and/or entire
folders into this panel. You can also right-click in the File list panel
to bring up a file selection window to browse for individual files; on
the Mac, folders can be drag-and-dropped from this window and you can
select multiple files using Ctrl-A (Windows) or Cmd-A (Mac).

.. image:: {IMG_PANEL_FILLED}
   :width: 100%

Right-clicking on the file list panel will provide a context menu with
options to modify the file list:

-  *Show Selected Image:* Selecting this option (or double-clicking on
   the file) will open the image in a new window.
-  *Remove From List:* Removes the selected file or folder from the
   list. Note that this does not remove the file/folder from the hard
   drive.
-  *Remove Unavailable Files:* Refresh the list by checking for
   existence of file. Note that this does not remove the files from the
   hard drive.
-  *Browse For Images:* Use a dialog box to select an image file (though
   drag-and-drop is recommended).
-  *Refresh:* Shown only if folder is selected. Refresh the list of
   files from the folder. Files that were manually removed from the list
   for that folder are restored.
-  *Expand All Folders:* Expand all trees shown in the file list panel.
-  *Collapse All Folders:* Collapse all folder trees shown in the file
   list panel.
-  *Clear File List:* Remove all files/folders in the file list panel.
   You will be prompted for confirmation beforehand. Note that this does
   not remove the files from the hard drive.

How do I configure the module?
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

If you have a subset of files that you want to analyze from the full
listing shown in the panel, you can filter the files according to a set
of rules. This is useful in cases such as:

-  You have dragged a folder of images onto the file list panel, but the
   folder contains images you want to analyze along with non-image files
   that you want to disregard.
-  You have dragged a folder of images onto the file list panel, but the
   folder contains the images from one experiment that you want to
   process along with images from another experiment that you want to
   ignore for now.

You may specify as many rules as necessary to define the desired list of
images.

After you have filtered the file list, press the “Apply” button to
update the view of the file list. You can also toggle the “Show file
excluded by filters” box to modify the display of the files:

-  Checking this box will show all the files in the list, with the files
   that have been filtered out shown as grayed-out entries.
-  Not checking this box will only show the files in the list that pass
   the filter(s).

What do I get as output?
^^^^^^^^^^^^^^^^^^^^^^^^

The final product of the **Images** module is a file list in which any
files that are not intended for further processing have been removed,
whether manually or using filtering. This list will be used when
collecting metadata (if desired) and when assembling the image sets in
NamesAndTypes. The list can be filtered further in NamesAndTypes to
specify, for example, that a subset of these images represents a
particular wavelength.

.. _here: http://www.openmicroscopy.org/site/support/bio-formats5/supported-formats.html
.. _“lossy”: http://www.techterms.com/definition/lossy
.. _“lossless”: http://www.techterms.com/definition/lossless
""".format(
    **{
        "IMG_PANEL_BLANK": image_resource("Images_FilelistPanel_Blank.png"),
        "IMG_PANEL_FILLED": image_resource("Images_FilelistPanel_Filled.png"),
    }
)


class Images(Module):
    variable_revision_number = 3
    module_name = "Images"
    category = "File Processing"

    MI_SHOW_IMAGE = "Show image"
    MI_REMOVE = FileCollectionDisplay.DeleteMenuItem("Remove from list")
    MI_REFRESH = "Refresh"

    def create_settings(self):
        self.workspace = None
        module_explanation = f"To begin creating your project, use the {self.module_name} module to compile a list " \
                             f"of files and/or folders that you want to analyze. You can also specify a set of rules " \
                             f"to include only the desired files in your selected folders."
        self.set_notes(module_explanation)

        self.path_list_display = PathListDisplay()
        predicates = [FilePredicate(), DirectoryPredicate(), ExtensionPredicate()]

        self.filter_choice = Choice(
            "Filter images?",
            FILTER_CHOICE_ALL,
            value=FILTER_CHOICE_IMAGES,
            doc=f"""\
The **Images** module will pass all the files specified in the file list
panel downstream to have a meaningful name assigned to it (so other
modules can access it) or optionally, to define the relationships
between images and associated metadata. Enabling file filtering will
allow you to specify a subset of the files from the file list panel by
defining rules to filter the files. This approach is useful if, for
example, you drag-and-dropped a folder onto the file list panel which
contains a mixture of images that you want to analyze and other files
that you want to ignore.

Several options are available for this setting:

-  *{FILTER_CHOICE_NONE}:* Do not enable filtering; all files in the
   File list panel will be passed to downstream modules for processing.
   This option can be selected if you are sure that only images are
   specified in the list.
-  *{FILTER_CHOICE_IMAGES}:* Only image files will be passed to
   downstream modules. The permissible image formats are provided by a
   library called Bio-Formats; see `here`_ for the formats available.
-  *{FILTER_CHOICE_CUSTOM}:* Specify custom rules for selecting a
   subset of the files from the File list panel. This approach is useful
   if, for example, you drag-and-dropped a folder onto the File list
   panel which contains a mixture of images that you want to analyze and
   other files that you want to ignore.

.. _here: http://www.openmicroscopy.org/site/support/bio-formats5/supported-formats.html
""",
        )

        self.filter = Filter(
            "Select the rule criteria",
            predicates,
            FILTER_DEFAULT,
            doc="""\
Specify a set of rules to narrow down the files to be analyzed.

{FILTER_RULES_BUTTONS_HELP}
""".format(
                **{"FILTER_RULES_BUTTONS_HELP": FILTER_RULES_BUTTONS_HELP}
            ),
        )

        self.update_button = PathListRefreshButton(
            "Apply filters to the file list",
            "Apply filters to the file list",
            doc="""\
*(Only displayed if filtering based on rules)*

Re-display the file list, removing or graying out the files that do not
pass the current filter.
""",
        )

        self.want_split = Binary("Extract image planes", False,
                                 doc="Choose whether to search for multiple images within files")
        self.split_C = Binary("Split channels", True,
                              doc="Choose whether to split color images into greyscale channels")
        self.split_Z = Binary("Split z-planes", True,
                              doc="Choose whether to split z-stacks into individual planes")
        self.split_T = Binary("Split timepoints", True,
                              doc="Choose whether to split time series into individual planes")

        self.extract_metadata = PathListExtractButton(
            "Run extraction", "Scan files",
            doc="Perform extraction to identify multiple images within a single file.")

    def help_settings(self):
        return [self.filter_choice, self.filter, self.update_button,
                self.want_split, self.split_C, self.split_Z, self.split_T, self.extract_metadata]


    def settings(self):
        return [self.path_list_display, self.filter_choice, self.filter,
                self.want_split, self.split_C, self.split_Z, self.split_T]

    def visible_settings(self):
        result = [self.path_list_display, self.filter_choice]
        if self.filter_choice == FILTER_CHOICE_CUSTOM:
            result += [self.filter, self.update_button]
            self.path_list_display.using_filter = True
        elif self.filter_choice == FILTER_CHOICE_IMAGES:
            result += [self.update_button]
            self.path_list_display.using_filter = True
        else:
            self.path_list_display.using_filter = False
        result += [self.want_split]
        if self.want_split.value:
            result += [self.split_C, self.split_Z, self.split_T, self.extract_metadata]
        return result

    def change_causes_prepare_run(self, setting):
        """Return True if a change to the settings requires a call to prepare_run

        Images should return True if any setting changes because that
        will affect the image plane descriptors passed onto later modules
        which will change the image set produced by the pipeline.
        """
        return setting in self.settings()

    @classmethod
    def is_input_module(self):
        return True

    def on_activated(self, workspace):
        self.pipeline = workspace.pipeline

    def filter_file_list(self, workspace):
        file_list = workspace.pipeline.file_list
        if file_list and not workspace.file_list.has_files():
            LOGGER.warning("Workspace file list is empty, will populate from pipeline."
                            "This may happen if you're running in headless mode.")
            workspace.file_list.add_files_to_filelist([f.url for f in file_list])
        if self.filter_choice != FILTER_CHOICE_NONE:
            if self.filter_choice == FILTER_CHOICE_IMAGES:
                file_list = [image_file for image_file in file_list if is_image(image_file.url)]
            else:
                new_file_list = []
                for image_file in file_list:
                    modpath = image_file.modpath
                    if self.filter.evaluate(
                        (FileCollectionDisplay.NODE_FILE, modpath, None,)
                    ):
                        new_file_list.append(image_file)
                file_list = new_file_list
        workspace.pipeline.set_filtered_file_list(file_list, self)
        return file_list

    def prepare_run(self, workspace):
        """Create an IPD for every url that passes the filter"""
        if workspace.pipeline.in_batch_mode():
            return True
        file_list = self.filter_file_list(workspace)
        if self.want_split.value:
            LOGGER.debug("Metadata extraction will be performed now if needed")
            if self.extract_metadata.callback is not None:
                # If GUI is present perform extraction and refresh the file list GUI
                self.extract_metadata.callback()
            else:
                # Otherwise, just do the extraction here
                for file_object in file_list:
                    if not file_object.extracted:
                        file_object.extract_planes(workspace)

        planes = []
        for image_file in file_list:
            if self.want_split:
                for seriesIdx, sizeC, sizeZ, sizeT, sizeY, sizeX, seriesName in image_file.get_plane_iterator():
                    to_split = [range(sizeC) if self.split_C.value else [None],
                                range(sizeZ) if self.split_Z.value else [None],
                                range(sizeT) if self.split_T.value else [None]]
                    is_color = sizeC > 1 and not self.split_C.value

                    for C, Z, T in itertools.product(*to_split):
                        planes.append(ImagePlane(image_file, series=seriesIdx, channel=C, z=Z, t=T, color=is_color,
                                                 name=seriesName))
            else:
                planes.append(ImagePlane(image_file))
        workspace.pipeline.set_image_plane_list(planes)
        return True

    def run(self, workspace):
        pass

    def get_metadata_keys(self):
        """Return a collection of metadata keys to be associated with files"""
        if not self.want_split.value:
            return []
        result = [C_SERIES, C_SERIES_NAME]
        if self.split_C.value:
            result.append(C_C)
        if self.split_T.value:
            result.append(C_T)
        if self.split_Z.value:
            result.append(C_Z)
        return result

    def upgrade_settings(self, setting_values, variable_revision_number, module_name):
        """Upgrade pipeline settings from a previous revision

        setting_values - the text values of the module's settings

        variable_revision_number - revision # of module version that saved them

        Returns upgraded setting values, revision number and matlab flag
        """
        if variable_revision_number == 1:
            # Changed from yes/no for filter to a choice
            filter_choice = (
                FILTER_CHOICE_CUSTOM
                if setting_values[1] == "Yes"
                else FILTER_CHOICE_NONE
            )
            setting_values = setting_values[:1] + [filter_choice] + setting_values[2:]
            variable_revision_number = 2
        return setting_values, variable_revision_number

    def volumetric(self):
        return True



import skimage.color

from ._url_image import URLImage


class ColorImage(URLImage):
    """Provide a color image, tripling a monochrome plane if needed"""

    def __init__(
        self, name, url, series, index, rescale_range=None, metadata_rescale=False, volume=False, spacing=None, z=None, t=None
    ):
        URLImage.__init__(
            self,
            name,
            url,
            rescale_range=rescale_range,
            metadata_rescale=metadata_rescale,
            series=series,
            index=index,
            volume=volume,
            spacing=spacing,
            z=z,
            t=t
        )

    def provide_image(self, image_set):
        image = URLImage.provide_image(self, image_set)

        if image.pixel_data.ndim == image.dimensions:
            image.pixel_data = skimage.color.gray2rgb(image.pixel_data)

        return image



import bisect
import re

import numpy

from ...constants.pipeline import make_cell_struct_dtype
from ...constants.pipeline import PIPELINE
from ...constants.pipeline import IMAGE_PLANE_DESCRIPTOR_VERSION
from ...constants.pipeline import H_PLANE_COUNT
from ...constants.pipeline import H_URL
from ...constants.pipeline import H_SERIES
from ...constants.pipeline import H_INDEX
from ...constants.pipeline import H_CHANNEL
from ...constants.pipeline import RF_STATE_PREQUOTE
from ...constants.pipeline import RF_STATE_FIELD
from ...constants.pipeline import RF_STATE_BACKSLASH_ESCAPE
from ...constants.pipeline import RF_STATE_SEPARATOR
from ..legacy import cmp


def add_all_images(handles, image_set, object_set):
    """ Add all images to the handles structure passed

    Add images to the handles structure, for example in the Python sandwich.
    """
    images = {}
    for provider in image_set.providers:
        name = provider.name()
        image = image_set.get_image(name)
        images[name] = image.image
        if image.has_mask:
            images["CropMask" + name] = image.mask

    for object_name in object_set.object_names:
        objects = object_set.get_objects(object_name)
        images["Segmented" + object_name] = objects.segmented
        if objects.has_unedited_segmented():
            images["UneditedSegmented" + object_name] = objects.unedited_segmented
        if objects.has_small_removed_segmented():
            images[
                "SmallRemovedSegmented" + object_name
            ] = objects.small_removed_segmented

    npy_images = numpy.ndarray(
        (1, 1), dtype=make_cell_struct_dtype(list(images.keys()))
    )
    for key, image in list(images.items()):
        npy_images[key][0, 0] = image
    handles[PIPELINE] = npy_images


def map_feature_names(feature_names, max_size=63):
    """Map feature names to legal Matlab field names

    returns a dictionary where the key is the field name and
    the value is the feature name.
    """
    mapping = {}
    seeded = False

    # def shortest_first(a, b):
    #     return (
    #         -1
    #         if len(a) < len(b)
    #         else 1
    #         if len(b) < len(a)
    #         else cellprofiler_core.utilities.legacy.cmp(a, b)
    #     )

    for feature_name in feature_names:
        # if len(feature_name) > max_size:
        #     name = feature_name
        #     to_remove = len(feature_name) - max_size
        #     remove_count = 0
        #     for to_drop in (
        #         ("a", "e", "i", "o", "u"),
        #         (
        #             "b",
        #             "c",
        #             "d",
        #             "f",
        #             "g",
        #             "h",
        #             "j",
        #             "k",
        #             "l",
        #             "m",
        #             "n",
        #             "p",
        #             "q",
        #             "r",
        #             "s",
        #             "t",
        #             "v",
        #             "w",
        #             "x",
        #             "y",
        #             "z",
        #         ),
        #         (
        #             "A",
        #             "B",
        #             "C",
        #             "D",
        #             "E",
        #             "F",
        #             "G",
        #             "H",
        #             "I",
        #             "J",
        #             "K",
        #             "L",
        #             "M",
        #             "N",
        #             "O",
        #             "P",
        #             "Q",
        #             "R",
        #             "S",
        #             "T",
        #             "U",
        #             "V",
        #             "W",
        #             "X",
        #             "Y",
        #             "Z",
        #         ),
        #     ):
        #         for index in range(len(name) - 1, -1, -1):
        #             if name[index] in to_drop:
        #                 name = name[:index] + name[index + 1 :]
        #                 remove_count += 1
        #                 if remove_count == to_remove:
        #                     break
        #         if remove_count == to_remove:
        #             break
        #     if name in list(mapping.keys()) or len(name) > max_size:
        #         # Panic mode - a duplication
        #         if not seeded:
        #             numpy.random.seed(0)
        #             seeded = True
        #         while True:
        #             npname = numpy.fromstring(feature_name, "|S1")
        #             indices = numpy.random.permutation(len(name))[:max_size]
        #             indices.sort()
        #             name = npname[indices]
        #             name = name.tostring()
        #             if not name in list(mapping.keys()):
        #                 break
        # else:
        #     name = feature_name
        mapping[feature_name] = feature_name
    return mapping


def read_file_list(file_or_fd):
    """Read a file list from a file or file object

    file_or_fd - either a path string or a file-like object

    Returns a collection of urls

    The unicode text for each field is utf-8 encoded, then string_escape encoded.
    All fields are delimited by quotes and separated by commas. A "None" value is
    represented by two consecutive commas. The series, index, and channel
    are no longer set, so the URL is followed by three commas.

    There are two header lines. The first header line has key/value pairs.
    Required key/value pairs are "Version" and "PlaneCount". "Version" is
    the format version. "PlaneCount" is the number of image planes to be
    read or saved.

    The second header line is legacy at this point - it used to specify
    a particular image plane in the file, but doesn't do that any longer.

    A minimal example:

    "Version":"1","PlaneCount":"1"
    "URL","Series","Index","Channel"
    "file:///imaging/analysis/singleplane.tif",,,
    """

    if isinstance(file_or_fd, str):
        needs_close = True
        fd = open(file_or_fd, "r")
    else:
        needs_close = False
        fd = file_or_fd
    try:
        line = next(fd)
        properties = dict(read_fields(line))
        if "Version" not in properties:
            raise ValueError("Image plane details header is missing its version #")
        version = int(properties["Version"])
        if version != IMAGE_PLANE_DESCRIPTOR_VERSION:
            raise ValueError(
                "Unable to read image plane details version # %d" % version
            )
        plane_count = int(properties[H_PLANE_COUNT])
        header = read_fields(next(fd))
        result = []
        pattern = r'(?:"((?:[^\\]|\\.)+?)")?(?:,|\s+)'
        for i in range(plane_count):
            fields = [x.groups()[0] for x in re.finditer(pattern, next(fd))]
            fields = [None if x is None else x for x in fields]
            url = fields[0]
            result.append(url)
        return result
    finally:
        if needs_close:
            fd.close()


def write_file_list(file_or_fd, file_list):
    """Write the file list out to a file.

    See read_image_plane_details for the file format.

    file_or_fd - a path or a file like object

    file_list - collection of URLs to be output

    """
    if isinstance(file_or_fd, str):
        fd = open(file_or_fd, "w")
        needs_close = True
    else:
        fd = file_or_fd
        needs_close = False
    try:
        fd.write(
            '"%s":"%d","%s":"%d"\n'
            % ("Version", IMAGE_PLANE_DESCRIPTOR_VERSION, H_PLANE_COUNT, len(file_list))
        )
        fd.write('"' + '","'.join([H_URL, H_SERIES, H_INDEX, H_CHANNEL]) + '"\n')
        for url in file_list:
            if isinstance(url, str):
                url = url
            # url = url.encode("string_escape").replace('"', r"\"")
            line = '"%s",,,\n' % url
            fd.write(line)
    finally:
        if needs_close:
            fd.close()


def read_fields(line):
    state = RF_STATE_PREQUOTE
    kv = False
    result = []
    field = ""
    for c in line:
        if state == RF_STATE_PREQUOTE:
            if c == '"':
                state = RF_STATE_FIELD
                field = ""
            elif c == ":":
                key = None
                kv = True
            elif c in ",\n":
                if kv:
                    result.append((key, None))
                else:
                    result.append(None)
                kv = False
        elif state == RF_STATE_BACKSLASH_ESCAPE:
            field += c
            state = RF_STATE_FIELD
        elif state == RF_STATE_SEPARATOR:
            if c == ":":
                key = field
                kv = True
                state = RF_STATE_PREQUOTE
            elif c in ",\n":
                field = field
                if kv:
                    result.append((key, field))
                else:
                    result.append(field)
                kv = False
                state = RF_STATE_PREQUOTE
        elif c == "\\":
            field += c
            state = RF_STATE_BACKSLASH_ESCAPE
        elif c == '"':
            state = RF_STATE_SEPARATOR
        else:
            field += c
    return result


def find_image_plane_details(exemplar, ipds):
    """Find the ImagePlaneDetails instance matching the exemplar

    The point of this function is to retrieve the ImagePlaneDetails from
    the list provided and, in doing so, get the attached metadata and the
    Java IPD object as well.

    exemplar - an IPD with the URL, series, index and channel filled in

    ipds - an ordered list of ImagePlaneDetails instances

    Returns the match or None if not found
    """
    pos = bisect.bisect_left(ipds, exemplar)
    if pos == len(ipds) or cmp(ipds[pos], exemplar):
        return None
    return ipds[pos]


def new_string_cell_array(shape):
    """Return a numpy.ndarray that looks like {NxM cell} to Matlab

    Return a numpy.ndarray that looks like {NxM cell} to Matlab.
    Each of the cells looks empty.
    shape - the shape of the array that's generated, e.g., (5,19) for a 5x19 cell array.
            Currently, this must be a 2-d shape.
    The object returned is a numpy.ndarray with dtype=dtype('object') and the given shape
    with each cell in the array filled with a numpy.ndarray with shape = (1,0)
    and dtype=dtype('float64'). This appears to be the form that's created in matlab
    for this sort of object.
    """
    result = numpy.ndarray(shape, dtype=numpy.dtype("object"))
    for i in range(0, shape[0]):
        for j in range(0, shape[1]):
            result[i, j] = numpy.empty((0, 0))
    return result


def encapsulate_strings_in_arrays(handles):
    """Recursively descend through the handles structure, replacing strings as arrays packed with strings

    This function makes the handles structure loaded through the sandwich compatible with loadmat. It operates on the array in-place.
    """
    if handles.dtype.kind == "O":
        # cells - descend recursively
        flat = handles.flat
        for i in range(0, len(flat)):
            if isinstance(flat[i], str):
                flat[i] = encapsulate_string(flat[i])
            elif isinstance(flat[i], numpy.ndarray):
                encapsulate_strings_in_arrays(flat[i])
    elif handles.dtype.fields:
        # A structure: iterate over all structure elements.
        for field in list(handles.dtype.fields.keys()):
            if isinstance(handles[field], str):
                handles[field] = encapsulate_string(handles[field])
            elif isinstance(handles[field], numpy.ndarray):
                encapsulate_strings_in_arrays(handles[field])


def encapsulate_string(s):
    """Encapsulate a string in an array of shape 1 of the length of the string
    """
    result = numpy.ndarray((1,), "<U%d" % (len(s)))

    result[0] = s

    return result



from .._subscriber import Subscriber


class ImageSubscriber(Subscriber):
    """A setting that provides an image name
    """

    def __init__(
        self,
        text,
        value=None,
        can_be_blank=False,
        blank_text="Leave blank",
        *args,
        **kwargs,
    ):
        super(ImageSubscriber, self).__init__(
            text, "imagegroup", value, can_be_blank, blank_text, *args, **kwargs
        )



def filter_duplicate_names(name_list):
    """remove any repeated names from a list of (name, ...) keeping the last occurrence."""
    name_dict = dict(list(zip((n[0] for n in name_list), name_list)))

    return [name_dict[n[0]] for n in name_list]


def get_name_providers(pipeline, last_setting):
    """Scan the pipeline to find name providers matching the name given in the setting
    pipeline - pipeline to scan
    last_setting - scan the modules in order until you arrive at this setting
    returns a list of providers that provide a correct "thing" with the
    same name as that of the subscriber
    """
    from ..setting.text.alphanumeric.name import Name

    choices = []

    for module in pipeline.modules(False):
        module_choices = []

        for setting in module.visible_settings():
            if setting.key() == last_setting.key():
                return choices

            if (
                isinstance(setting, Name)
                and setting != "Do not use"
                and module.enabled
                and last_setting.matches(setting)
                and setting.value == last_setting.value
            ):
                module_choices.append(setting)

        choices += module_choices

    assert False, "Setting not among visible settings in pipeline"


def get_name_provider_choices(pipeline, last_setting, group):
    """Scan the pipeline to find name providers for the given group

    pipeline - pipeline to scan
    last_setting - scan the modules in order until you arrive at this setting
    group - the name of the group of providers to scan
    returns a list of tuples, each with (provider name, module name, module number)
    """
    from ..setting.text.alphanumeric.name import Name

    choices = []

    for module in pipeline.modules(False):
        module_choices = [
            (
                other_name,
                module.module_name,
                module.module_num,
                module.is_input_module(),
            )
            for other_name in module.other_providers(group)
        ]

        for setting in module.visible_settings():
            if setting.key() == last_setting.key():
                return filter_duplicate_names(choices)

            if (
                isinstance(setting, Name)
                and module.enabled
                and setting != "Do not use"
                and last_setting.matches(setting)
            ):
                module_choices.append(
                    (
                        setting.value,
                        module.module_name,
                        module.module_num,
                        module.is_input_module(),
                    )
                )

        choices += module_choices

    assert False, "Setting not among visible settings in pipeline"



import collections

from cellprofiler_core.pipeline import ImagePlane


class ChannelHasher:
    """
    This class is used by NamesAndTypes to pair image sets by metadata.
    It forms a wrapper around a defaultdict object which returns lists.
    Those lists will store ImagePlane objects matching a set of
    metadata values.

    channel_name is the name of the image channel.

    keymap is a list of metadata keys to be used to index the ImagePlane objects.

    An ImagePlane is assigned a hash key tuple based on the keymap parameter.
    Sometimes a channel doesn't use all metadata keys, in which case the unused
    key should be passed as None within the keymap. All ChannelHasher objects
    in use should have a key list of the same length.

    e.g. If we're using Plate, Well and Site keys to match metadata,
    ['Plate', 'Well', 'Site'] should be the keymap for the channel. When adding
    planes to the channel group they'll be indexed with a tuple of (Plate, Well, Site)
    values. For successful image set creation each hash key should only end up with a
    single ImagePlane.

    If we have an Illum function which is assigned based on just the Plate key, the
    keymap should be ['Plate', None, None]. Internally the hasher will call
    for just (Plate) when NamesAndTypes calls for Plate X, Well Y, Site Z. This
    allows a single image to match in multiple sets.
    """
    def __init__(self, channel_name, keymap):
        self.name = channel_name
        self.mapper = collections.defaultdict(list)
        self.keymap = keymap
        self.hash_keys = [key for key in keymap if key is not None]

    def add(self, image_plane: ImagePlane):
        plane_hash = tuple([image_plane.get_metadata(key) for key in self.hash_keys])
        self.mapper[plane_hash].append(image_plane)

    def keys(self):
        return self.mapper.keys()

    def __getitem__(self, key):
        real_key = []
        for key, mapped_key in zip(key, self.keymap):
            if mapped_key is None:
                continue
            real_key.append(key)
        return self.mapper[tuple(real_key)]

    def __setitem__(self, key, value):
        if len(key) == len(self.hash_keys):
            self.mapper[key].append(value)
        else:
            real_key = []
            for key, mapped_key in zip(key, self.keymap):
                if mapped_key is None:
                    continue
                real_key.append(key)
            self.mapper[tuple(real_key)] = value



class SettingsGroup:
    """A group of settings that are managed together in the UI.
    Particulary useful when used with a RemoveSettingButton.
    Individual settings can be added with append(), and their value
    fetched from the group using the name given in append.
    """

    def __init__(self):
        self.settings = []

    def append(self, name, setting):
        """Add a new setting to the group, with a name.  The setting
        will then be available as group.name
        """
        assert name not in self.__dict__, (
            "%s already in SettingsGroup (previous setting or built in attribute)"
            % name
        )
        self.__setattr__(name, setting)
        self.settings.append(setting)

    def visible_settings(self):
        """Return a list of the settings in the group, in the order
        they were added to the group.
        """
        # return a copy
        return list(self.settings)

    def pipeline_settings(self):
        """Return a list of the settings, filtering out UI tidbits"""
        return [setting for setting in self.settings if setting.save_to_pipeline]



class MetadataGroup(dict):
    """A set of metadata tag values and the image set indexes that match

    The MetadataGroup object represents a group of image sets that
    have the same values for a given set of tags. For instance, if an
    experiment has metadata tags of "Plate", "Well" and "Site" and
    we form a metadata group of "Plate" and "Well", then each metadata
    group will have image set indexes of the images taken of a particular
    well
    """

    def __init__(self, tag_dictionary, image_numbers):
        super(MetadataGroup, self).__init__(tag_dictionary)
        self.__image_numbers = image_numbers

    @property
    def image_numbers(self):
        return self.__image_numbers

    def __setitem__(self, tag, value):
        raise NotImplementedError("The dictionary is read-only")



from ...utilities.zmq.communicable.reply import Reply


class ExceptionPleaseDebug(Reply):
    def __init__(self, disposition, verification_hash=None):
        Reply.__init__(
            self, disposition=disposition, verification_hash=verification_hash
        )



"""
/!\ WARNING - DEPRECATED /!\
utf16encode.py - encode unicode strings as escaped utf16
This is only used for pipeline version < 3 files
"""


def utf16decode(x):
    """Decode an escaped utf8-encoded string
    """
    y = ""
    state = -1
    for z in x:
        if state == -1:
            if z == "\\":
                state = 0
            else:
                y += str(z)
        elif state == 0:
            if z == "u":
                state = 1
                acc = ""
            else:
                y += str(z)
                state = -1
        elif state < 4:
            state += 1
            acc += z
        else:
            state = -1
            y += chr(int(acc + z, 16))
    return y



class Listener:
    """A class to wrap add/remove listener for use with "with"

    Usage:
    def my_listener(pipeline, event):
        .....
    with pipeline.PipelineListener(pipeline, my_listener):
        # listener has been added
        .....
    # listener has been removed
    """

    def __init__(self, pipeline, listener):
        self.pipeline = pipeline
        self.listener = listener

    def __enter__(self):
        self.pipeline.add_listener(self.listener)
        return self

    def __exit__(self, exc_type, exc_value, tb):
        self.pipeline.remove_listener(self.listener)



import click

from ..__main__ import pass_environment


@click.group("pipeline")
@pass_environment
def command(context):
    pass


@command.command("measurements", help="returns measurements extracted by the pipeline")
@pass_environment
def measurements(context):
    pass


@command.command("run", help="executes the pipeline")
@click.argument("pipeline", type=click.File("r"))
@click.option("--batch-size", type=int)
@click.option("--data", type=click.Path())
@click.option("--default-images-directory", type=click.Path())
@click.option("--default-output-directory", type=click.Path())
@click.option("--images", multiple=True, type=click.Path())
@click.option("--beginning", default=1, type=int)
@click.option("--end", type=int)
@click.option("--group", type=str)
@pass_environment
def run(
    context,
    batch_size,
    data,
    images_directory,
    output_directory,
    images,
    beginning,
    end,
    grouping,
):
    pass



class Dependency:
    """This class documents the dependency of one module on another

    A module is dependent on another if the dependent module requires
    data from the producer module. That data can be objects (label matrices),
    a derived image or measurements.
    """

    def __init__(
        self,
        source_module,
        destination_module,
        source_setting=None,
        destination_setting=None,
    ):
        """Constructor

        source_module - the module that produces the data
        destination_module - the module that uses the data
        source_setting - the module setting that names the item (can be None)
        destination_setting - the module setting in the destination that
        picks the setting
        """
        self.__source_module = source_module
        self.__destination_module = destination_module
        self.__source_setting = source_setting
        self.__destination_setting = destination_setting

    @property
    def source(self):
        """The source of the data item"""
        return self.__source_module

    @property
    def source_setting(self):
        """The setting that names the data item

        This can be None if it's ambiguous.
        """
        return self.__source_setting

    @property
    def destination(self):
        """The user of the data item"""
        return self.__destination_module

    @property
    def destination_setting(self):
        """The setting that picks the data item

        This can be None if it's ambiguous.
        """
        return self.__destination_setting



from ._file_image import FileImage


class FlexFrameImage(FileImage):
    """Provide an image by filename:frame, loading the file as it is requested
    """

    def __init__(self,
                 name,
                 pathname,
                 filename,
                 series,
                 index,
                 rescale_range=None,
                 metadata_rescale=False
                ):
        super(FlexFrameImage, self).__init__(
            name,
            pathname,
            filename,
            rescale_range=rescale_range,
            metadata_rescale=metadata_rescale,
            series=series,
            index=index
        )



import sys

import cellprofiler_core.utilities.zmq


class Communicable:
    """Base class for Requests and Replies.

    All subclasses must accept keyword arguments to __init__() corresponding to
    their attributes.
    """

    def send(self, socket, routing=None):
        if routing is None:
            routing = []
        if hasattr(self, "_remote"):
            assert not self._remote, "send() called on a non-local Communicable object."
        json_str, buffers = cellprofiler_core.utilities.zmq.json_encode(self.__dict__)
        json_str = json_str.encode("utf-8")
        message_parts = (
            routing
            + [
                self.__class__.__module__.encode("utf-8"),
                self.__class__.__name__.encode("utf-8"),
            ]
            + [json_str]
        )
        socket.send_multipart(
            message_parts + buffers, copy=False,
        )

    class MultipleReply(RuntimeError):
        pass

    @classmethod
    def recv(cls, socket, routed=False):
        message = socket.recv_multipart()
        if routed:
            split = message.index(b"") + 1
            routing = message[:split]
            message = message[split:]
        else:
            routing = []
        module, classname = message[:2]
        module = module.decode("unicode_escape")
        classname = classname.decode("unicode_escape")
        buffers = message[3:]
        attribute_dict = cellprofiler_core.utilities.zmq.json_decode(
            message[2], buffers
        )
        try:
            instance = sys.modules[module].__dict__[classname](**attribute_dict)
        except:
            print(
                "Communicable could not instantiate %s from module %s with kwargs %s"
                % (module, classname, attribute_dict)
            )
            raise
        instance._remote = True
        instance._routing = routing
        instance._socket = socket
        instance._replied = False
        return instance

    def routing(self):
        return "/".join(self._routing)

    def reply(self, reply_obj, please_reply=False):
        assert self._remote, "Replying to a local Communicable!"
        if self._replied:
            raise self.MultipleReply("Can't reply to a Communicable more than once!")
        Communicable.send(reply_obj, self._socket, self._routing)
        self._replied = True
        if please_reply:
            raise NotImplementedError(
                "Req / rep / rep / rep pattern is no longer supported"
            )



from ._integer_range import IntegerRange
from .._range import Range
from ..._validation_error import ValidationError


class IntegerOrUnboundedRange(IntegerRange):
    """A setting that specifies an integer range where the minimum and maximum
    can be set to unbounded by the user.

    The maximum value can be relative to the far side in which case a negative
    number is returned for slicing.
    """

    def __init__(
        self, text, value=(0, "end"), minval=None, maxval=None, *args, **kwargs
    ):
        """Initialize an integer range
        text  - helpful text to be displayed to the user
        value - initial default value, a two-tuple as minimum and maximum
        minval - the minimum acceptable value of either
        maxval - the maximum acceptable value of either
        """
        Range.__init__(
            self, text, "%s,%s" % (str(value[0]), str(value[1])), *args, **kwargs
        )

    def str_to_value(self, str_value):
        if str_value == "begin":
            return 0
        elif (self.is_abs() and str_value == "end") or (
            len(str_value) > 0 and str_value[1:] == "end"
        ):
            return "end"
        return super(IntegerOrUnboundedRange, self).str_to_value(str_value)

    def value_to_str(self, value):
        if value in ("begin", "end"):
            return value
        return super(IntegerOrUnboundedRange, self).value_to_str(value)

    def get_unbounded_min(self):
        """True if there is no minimum"""
        return self.get_min() == 0

    unbounded_min = property(get_unbounded_min)

    def get_display_min(self):
        """What to display for the minimum"""
        return self.get_min_text()

    display_min = property(get_display_min)

    def get_unbounded_max(self):
        """True if there is no maximum"""
        return self.get_max_text() == "end"

    unbounded_max = property(get_unbounded_max)

    def get_display_max(self):
        """What to display for the maximum"""
        #
        # Remove the minus sign
        #
        mt = self.get_max_text()
        if self.is_abs():
            return mt
        return mt[1:]

    display_max = property(get_display_max)

    def compose_display_max_text(self, dm_value):
        """Compose a value_text value for the setting given a max text value

        dm_value - the displayed text for the maximum of the range

        Returns a text value suitable for this setting that sets the
        maximum while keeping the minimum and abs/rel the same
        """
        if self.is_abs():
            return self.compose_max_text(dm_value)
        else:
            return self.compose_max_text("-" + dm_value)

    def is_abs(self):
        """Return True if the maximum is an absolute # of pixels

        Returns False if the # of pixels is relative to the right edge.
        """
        mt = self.get_max_text()
        return len(mt) == 0 or mt[0] != "-"

    def compose_abs(self):
        """Compose a text value that uses absolute upper bounds coordinates

        Return a text value for IntegerOrUnboundedRange that keeps the min
        and the max the same, but states that the max is the distance in pixels
        from the origin.
        """
        return self.compose_max_text(self.get_display_max())

    def compose_rel(self):
        """Compose a text value that uses relative upper bounds coordinates

        Return a text value for IntegerOrUnboundedRange that keeps the min
        and the max the same, but states that the max is the distance in pixels
        from the side of the image opposite the origin.
        """
        return self.compose_max_text("-" + self.get_display_max())

    def test_valid(self, pipeline):
        values = self.value_text.split(",")
        if len(values) < 2:
            raise ValidationError(
                "Minimum and maximum values must be separated by a comma", self
            )
        if len(values) > 2:
            raise ValidationError("Only two values allowed", self)
        if (not values[0].isdigit()) and values[0] != "begin":
            raise ValidationError("%s is not an integer" % (values[0]), self)
        if len(values[1]) == 0:
            raise ValidationError("The end value is blank", self)
        if not (
            values[1] == "end"
            or values[1].isdigit()
            or (
                values[1][0] == "-"
                and (values[1][1:].isdigit() or values[1][1:] == "end")
            )
        ):
            raise ValidationError(
                "%s is not an integer or %s" % (values[1], "end"), self
            )
        if (not self.unbounded_min) and self._minval and self._minval > self.min:
            raise ValidationError(
                "%s can't be less than %d" % (self.min_text, self._minval), self
            )
        if (not self.unbounded_max) and self._maxval and self._maxval < self.max:
            raise ValidationError(
                "%d can't be greater than %d" % (self.max, self._maxval), self
            )
        if (
            (not self.unbounded_min)
            and (not self.unbounded_max)
            and self.min > self.max > 0
        ):
            raise ValidationError("%d is greater than %d" % (self.min, self.max), self)



from ._dependency import Dependency


class MeasurementDependency(Dependency):
    """A dependency on a measurement"""

    def __init__(
        self,
        source_module,
        destination_module,
        object_name,
        feature,
        source_setting=None,
        destination_setting=None,
    ):
        """Initialize using source, destination and measurement

        source_module - module producing the measurement

        destination_module - module using the measurement

        object_name - the measurement is made on the objects with this name
        (or Image for image measurements)

        feature - the feature name for the measurement, for instance AreaShape_Area

        source_setting - the module setting that controls production of this
        measurement (very typically = None for no such thing)

        destination_setting - the module setting that chooses the measurement
        for the user of the data, for instance a MeasurementSetting
        """
        super(type(self), self).__init__(
            source_module, destination_module, source_setting, destination_setting
        )
        self.__object_name = object_name
        self.__feature = feature

    @property
    def object_name(self):
        """The objects / labels used when producing the measurement"""
        return self.__object_name

    @property
    def feature(self):
        """The name of the measurement"""
        return self.__feature

    def __str__(self):
        return "Measurement: %s.%s" % (self.object_name, self.feature)



import hashlib

import numpy

from ..image import Image
from ..module import Module
from ..setting.text import Name


class InjectImage(Module):
    """This module is intended for testing. It injects an image into the
    image set.
    """

    module_name = "InjectImage"
    variable_revision_number = 1

    def __init__(self, image_name, image, mask=None, release_image=False):
        """Initializer

        image_name - the name of the image to put into the image set

        image - either the pixel data for the image if adding the same
                image to every set or a list or tuple of pixel data,
                one per image set

        mask - None for no mask (default), a binary 2-d matrix if same mask
               for all image sets or a list or tuple of masks if one
               different mask per image set.
        """
        super(InjectImage, self).__init__()
        self.__image_name = image_name
        self.__image = image
        self.__mask = mask
        self.image_name = Name("Hardwired image name", "imagegroup", image_name)
        self.__release_image = release_image

    def settings(self):
        return [self.image_name]

    def visible_settings(self):
        return [self.image_name]

    def get_help(self):
        """Return help text for the module

        """
        raise NotImplementedError(
            "Please implement GetHelp in your derived module class"
        )

    def write_to_handles(self, handles):
        """Write out the module's state to the handles

        """

    def write_to_text(self, file):
        """Write the module's state, informally, to a text file
        """

    def prepare_run(self, workspace):
        digest = hashlib.md5()
        digest.update(numpy.ascontiguousarray(self.__image).data)

        workspace.measurements.add_measurement(
            "Image", "MD5Digest_%s" % self.__image_name, 1, image_set_number=1,
        )

        return True

    def run(self, workspace):
        """Run the module (abstract method)

        pipeline     - instance of CellProfiler.Pipeline for this run
        image_set    - the images in the image set being processed
        object_set   - the objects (labeled masks) in this image set
        measurements - the measurements for this run
        """
        if isinstance(self.__image, (tuple, list)):
            image = self.__image[workspace.image_set.image_number - 1]
        else:
            image = self.__image
        if isinstance(self.__mask, (tuple, list)):
            mask = self.__mask[workspace.image_set.image_number - 1]
        else:
            mask = self.__mask
        image = Image(image, mask)
        workspace.image_set.add(self.__image_name, image)

    def post_run(self, workspace):
        if self.__release_image:
            del self.__image
            del self.__mask

    def get_categories(self, pipeline, object_name):
        """Return the categories of measurements that this module produces

        object_name - return measurements made on this object (or 'Image' for image measurements)
        """
        return []

    def get_measurements(self, pipeline, object_name, category):
        """Return the measurements that this module produces

        object_name - return measurements made on this object (or 'Image' for image measurements)
        category - return measurements made in this category
        """
        return []

    def get_measurement_images(self, pipeline, object_name, category, measurement):
        """Return a list of image names used as a basis for a particular measure
        """
        return []

    def get_measurement_scales(
        self, pipeline, object_name, category, measurement, image_name
    ):
        """Return a list of scales (eg for texture) at which a measurement was taken
        """
        return []




ASSIGN_ALL = "All images"
ASSIGN_GUESS = "Try to guess image assignment"
ASSIGN_RULES = "Images matching rules"

LOAD_AS_GRAYSCALE_IMAGE = "Grayscale image"
LOAD_AS_COLOR_IMAGE = "Color image"
LOAD_AS_MASK = "Binary mask"
LOAD_AS_MASK_V5A = "Mask"
LOAD_AS_ILLUMINATION_FUNCTION = "Illumination function"
LOAD_AS_OBJECTS = "Objects"
LOAD_AS_ALL = [
    LOAD_AS_GRAYSCALE_IMAGE,
    LOAD_AS_COLOR_IMAGE,
    LOAD_AS_MASK,
    LOAD_AS_ILLUMINATION_FUNCTION,
    LOAD_AS_OBJECTS,
]

INTENSITY_RESCALING_BY_DATATYPE = "Image bit-depth"
INTENSITY_RESCALING_BY_METADATA = "Image metadata"
INTENSITY_MANUAL = "Manual"
INTENSITY_ALL = [
    INTENSITY_RESCALING_BY_DATATYPE,
    INTENSITY_RESCALING_BY_METADATA,
    INTENSITY_MANUAL,
]
MANUAL_INTENSITY_LABEL = "Maximum intensity"

RESCALING_HELP_TEXT = """\
This option determines how the image intensity should be rescaled from
0.0 – 1.0.

-  *{INTENSITY_RESCALING_BY_DATATYPE}:* Rescale the image to 0 – 1
   depending on the number of bits used to store the image
   (e.g. 255, 65535, etc.)
-  *{INTENSITY_RESCALING_BY_METADATA}:* Rescale the image intensity
   so that saturated values are rescaled to 1.0 by dividing all pixels
   in the image by the maximum possible intensity value allowed by the
   imaging hardware. Some image formats save the maximum possible
   intensity value along with the pixel data. For instance, a microscope
   might acquire images using a 12-bit A/D converter which outputs
   intensity values between zero and 4095, but stores the values in a
   field that can take values up to 65535.Choosing this setting will try its
   hardest to ensure that the intensity scaling value is the maximum
   allowed by the hardware, and not the maximum allowable by the file format.
   Although, this information is often unavailable, so may need to default
   to image datatype instead.
-  *{INTENSITY_MANUAL}:* Divide each pixel value by the value entered
   in the *{MANUAL_INTENSITY_LABEL}* setting. *{INTENSITY_MANUAL}*
   can be used to rescale an image whose maximum intensity metadata
   value is absent or incorrect, but is less than the value that would
   be supplied if *{INTENSITY_RESCALING_BY_DATATYPE}* were specified.

Please note that CellProfiler does not provide the option of loading the
image as the raw, unscaled values. If you wish to make measurements on
the unscaled image, use the **ImageMath** module to multiply the scaled
image by the actual image bit-depth.
""".format(
    **{
        "INTENSITY_RESCALING_BY_DATATYPE": INTENSITY_RESCALING_BY_DATATYPE,
        "INTENSITY_RESCALING_BY_METADATA": INTENSITY_RESCALING_BY_METADATA,
        "INTENSITY_MANUAL": INTENSITY_MANUAL,
        "MANUAL_INTENSITY_LABEL": MANUAL_INTENSITY_LABEL,
    }
)

MANUAL_RESCALE_HELP_TEXT = """\
*(Used only if “{INTENSITY_MANUAL}” is chosen)*

**NamesAndTypes** divides the pixel value, as read from the image file,
by this value to get the loaded image’s per-pixel intensity.
""".format(
    **{"INTENSITY_MANUAL": INTENSITY_MANUAL}
)

LOAD_AS_CHOICE_HELP_TEXT = """\
You can specify how these images should be treated:

-  *{LOAD_AS_GRAYSCALE_IMAGE}:* An image in which each pixel
   represents a single intensity value. Most of the modules in
   CellProfiler operate on images of this type.
   If this option is applied to a color image, the red, green and blue
   pixel intensities will be averaged to produce a single intensity
   value.
-  *{LOAD_AS_COLOR_IMAGE}:* An image in which each pixel represents a
   red, green and blue (RGB) triplet of intensity values OR which contains
   multiple individual grayscale channels. Please note
   that the object detection modules such as **IdentifyPrimaryObjects**
   expect a grayscale image, so if you want to identify objects, you
   should use the **ColorToGray** module in the analysis pipeline to
   split the color image into its component channels.
   You can use the **ColorToGray**'s *Combine* option after image loading
   to collapse the color channels to a single grayscale value if you don’t need
   CellProfiler to treat the image as color.
-  *{LOAD_AS_MASK}:* A *mask* is an image where some of the pixel
   intensity values are zero, and others are non-zero. The most common
   use for a mask is to exclude particular image regions from
   consideration. By applying a mask to another image, the portion of
   the image that overlaps with the non-zero regions of the mask are
   included. Those that overlap with the zeroed region are “hidden” and
   not included in downstream calculations. For this option, the input
   image should be a binary image, i.e, foreground is white, background
   is black. The module will convert any nonzero values to 1, if needed.
   You can use this option to load a foreground/background segmentation
   produced by the **Threshold** module or one of the **Identify** modules.
-  *{LOAD_AS_ILLUMINATION_FUNCTION}:* An *illumination correction
   function* is an image which has been generated for the purpose of
   correcting uneven illumination/lighting/shading or to reduce uneven
   background in images. Typically, is a file in the NumPy .npy format.
   See **CorrectIlluminationCalculate** and **CorrectIlluminationApply**
   for more details.
-  *{LOAD_AS_OBJECTS}:* Use this option if the input image is a label
   matrix and you want to obtain the objects that it defines. A label
   matrix is a grayscale or color image in which the connected regions
   share the same label, which defines how objects are represented in
   CellProfiler. The labels are integer values greater than or equal to
   0. The elements equal to 0 are the background, whereas the elements
   equal to 1 make up one object, the elements equal to 2 make up a
   second object, and so on. This option allows you to use the objects
   immediately without needing to insert an **Identify** module to
   extract them first. See **IdentifyPrimaryObjects** for more details.
   This option can load objects created by using the **ConvertObjectsToImage**
   module followed by the **SaveImages** module. Loaded objects can take two
   forms, with different considerations for each:

   -  *Non-overlapping* objects are stored as a label matrix. This
      matrix should be saved as grayscale rather than color.
   -  *Overlapping objects* are stored in a multi-frame TIF, each frame
      of which consists of a grayscale label matrix. The frames are
      constructed so that objects that overlap are placed in different
      frames. CellProfiler currently does not support saving of overlapping
      objects, so these can only be used within the pipeline.
""".format(
    **{
        "LOAD_AS_COLOR_IMAGE": LOAD_AS_COLOR_IMAGE,
        "LOAD_AS_GRAYSCALE_IMAGE": LOAD_AS_GRAYSCALE_IMAGE,
        "LOAD_AS_ILLUMINATION_FUNCTION": LOAD_AS_ILLUMINATION_FUNCTION,
        "LOAD_AS_MASK": LOAD_AS_MASK,
        "LOAD_AS_OBJECTS": LOAD_AS_OBJECTS,
    }
)

IDX_ASSIGNMENTS_COUNT_V2 = 5
IDX_ASSIGNMENTS_COUNT_V3 = 6
IDX_ASSIGNMENTS_COUNT_V5 = 6
IDX_ASSIGNMENTS_COUNT_V6 = 6
IDX_ASSIGNMENTS_COUNT_V7 = 6
IDX_ASSIGNMENTS_COUNT = 6

IDX_SINGLE_IMAGES_COUNT_V5 = 7
IDX_SINGLE_IMAGES_COUNT_V6 = 7
IDX_SINGLE_IMAGES_COUNT_V7 = 7
IDX_SINGLE_IMAGES_COUNT = 7

IDX_FIRST_ASSIGNMENT_V3 = 7
IDX_FIRST_ASSIGNMENT_V4 = 7
IDX_FIRST_ASSIGNMENT_V5 = 8
IDX_FIRST_ASSIGNMENT_V6 = 9
IDX_FIRST_ASSIGNMENT_V7 = 13
IDX_FIRST_ASSIGNMENT = 13

NUM_ASSIGNMENT_SETTINGS_V2 = 4
NUM_ASSIGNMENT_SETTINGS_V3 = 5
NUM_ASSIGNMENT_SETTINGS_V5 = 7
NUM_ASSIGNMENT_SETTINGS_V6 = 8
NUM_ASSIGNMENT_SETTINGS_V7 = 8
NUM_ASSIGNMENT_SETTINGS = 6

NUM_SINGLE_IMAGE_SETTINGS_V5 = 7
NUM_SINGLE_IMAGE_SETTINGS_V6 = 8
NUM_SINGLE_IMAGE_SETTINGS_V7 = 8
NUM_SINGLE_IMAGE_SETTINGS = 6


OFF_LOAD_AS_CHOICE_V5 = 3
OFF_LOAD_AS_CHOICE = 3

OFF_SI_LOAD_AS_CHOICE_V5 = 3
OFF_SI_LOAD_AS_CHOICE = 3

MATCH_BY_ORDER = "Order"
MATCH_BY_METADATA = "Metadata"

IMAGE_NAMES = ["DNA", "GFP", "Actin"]
OBJECT_NAMES = ["Cell", "Nucleus", "Cytoplasm", "Speckle"]

DEFAULT_MANUAL_RESCALE = 255

"""The experiment measurement that holds the ZLIB compression dictionary for image sets"""
M_IMAGE_SET_ZIP_DICTIONARY = "ImageSet_Zip_Dictionary"
"""The image measurement that holds the compressed image set"""
M_IMAGE_SET = "ImageSet_ImageSet"

# Image set error types
E_WRONG_LENGTH = "missing images"
E_MISSING = "no matches"
E_TOO_MANY = "too many matches"




from ._setting import Setting
from ._validation_error import ValidationError


class Joiner(Setting):
    """The joiner setting defines a joining condition between conceptual tables

    You might want to join several tables by specifying the columns that match
    each other or might want to join images in an image set by matching
    their metadata. The joiner takes a dictionary of lists of column names
    or metadata keys where the dictionary key holds the table or image name
    and the list of values holds the names of table columns or metadata keys.

    The joiner's value is, conceptually, a list of dictionaries where each
    dictionary in the list documents how to join one column or metadata key
    in one of the tables or images to the others.

    The conceptual value is a list of dictionaries of unicode string keys
    and values (or value = None). This can be encoded using str() and
    can be decoded using eval.
    """

    def __init__(self, text, value="[]", allow_none=True, **kwargs):
        """Initialize the joiner

        text - label to the left of the joiner

        value - "repr" done on the joiner's underlying structure which is
                a list of dictionaries

        allow_none - True (by default) to allow one of the entities to have
                     None for a join, indicating that it matches against
                     everything
        """
        super(self.__class__, self).__init__(text, value, **kwargs)
        self.entities = {}
        self.allow_none = allow_none

    def parse(self):
        """Parse the value into a list of dictionaries

        return a list of dictionaries where the key is the table or image name
        and the value is the column or metadata
        """
        return eval(self.value_text, {"__builtins__": None}, {})

    def default(self):
        """Concoct a default join as a guess if setting is uninitialized"""
        all_names = {}
        best_name = None
        best_count = 0
        for value_list in list(self.entities.values()):
            for value in value_list:
                if value in all_names:
                    all_names[value] += 1
                else:
                    all_names[value] = 1
                if best_count < all_names[value]:
                    best_count = all_names[value]
                    best_name = value
        if best_count == 0:
            return []
        else:
            return [
                dict(
                    [
                        (k, best_name if best_name in self.entities[k] else None)
                        for k in list(self.entities.keys())
                    ]
                )
            ]

    def build(self, dictionary_list):
        """Build a value from a list of dictionaries"""
        self.value = self.build_string(dictionary_list)

    @classmethod
    def build_string(cls, dictionary_list):
        return str(dictionary_list)

    def test_valid(self, pipeline):
        """Test the joiner setting to ensure that the join is supported

        """
        join = self.parse()
        if len(join) == 0:
            raise ValidationError(
                "This setting needs to be initialized by choosing items from each column",
                self,
            )
        for d in join:
            for column_name, value in list(d.items()):
                if column_name in self.entities and (
                    value not in self.entities[column_name] and value is not None
                ):
                    raise ValidationError(
                        "%s is not a valid choice for %s" % (value, column_name), self
                    )



from .._subscriber import Subscriber
from ..._validation_error import ValidationError


class ListSubscriber(Subscriber):
    """
    Stores name provider names as a list
    """

    def __init__(
        self,
        text,
        group,
        value=None,
        can_be_blank=False,
        blank_text="Leave blank",
        *args,
        **kwargs,
    ):
        if value is None:
            value = ""
        super(ListSubscriber, self).__init__(
            text, group, value, can_be_blank, blank_text, *args, **kwargs
        )
        self.value = value

    def get_value_text(self):
        """Convert the underlying list to a string"""
        return ", ".join(map(str, self._Setting__value))

    def __internal_set_value_text(self, value):
        self.set_value_text(value)

    def set_value_text(self, value):
        self._Setting__value = value

    value_text = property(get_value_text, __internal_set_value_text)

    def __internal_set_value(self, value):
        """Convert a saved string into a list"""
        if len(value) == 0:
            value = []
        else:
            value = value.split(", ")
        self._Setting__value = value

    def __internal_get_value(self):
        return self.get_value()

    value = property(__internal_get_value, __internal_set_value)

    def test_valid(self, pipeline):
        choices = self.get_choices(pipeline)
        if len(choices) == 0:
            raise ValidationError(
                "No prior instances of %s were defined" % self.group, self
            )
        for name in self.value:
            if name not in [c[0] for c in choices]:
                raise ValidationError(
                    "%s not in %s"
                    % (name, ", ".join(c[0] for c in self.get_choices(pipeline))),
                    self,
                )



from ._event import Event


class PipelineCleared(Event):
    """Indicates that all modules have been removed from the pipeline

    """

    def __init__(self):
        super(PipelineCleared, self).__init__(
            is_pipeline_modification=True, is_image_set_modification=True
        )

    def event_type(self):
        return "PipelineCleared"



import cellprofiler_core.setting
from ._does_not_predicate import DoesNotPredicate
from ._does_predicate import DoesPredicate
from ._filter import CONTAINS_PREDICATE
from ._filter import CONTAINS_REGEXP_PREDICATE
from ._filter import ENDSWITH_PREDICATE
from ._filter import EQ_PREDICATE
from ._filter import STARTS_WITH_PREDICATE
from ._filter_predicate import FilterPredicate


class FilePredicate(FilterPredicate):
    """A predicate that only filters files"""

    def __init__(self):
        subpredicates = (
            CONTAINS_PREDICATE,
            CONTAINS_REGEXP_PREDICATE,
            STARTS_WITH_PREDICATE,
            ENDSWITH_PREDICATE,
            EQ_PREDICATE,
        )

        predicates = [
            DoesPredicate(subpredicates),
            DoesNotPredicate(subpredicates),
        ]

        FilterPredicate.__init__(
            self,
            "file",
            "File",
            self.fn_filter,
            predicates,
            doc="Apply the rule to files",
        )

    @staticmethod
    def fn_filter(node_type__modpath__module, *args):
        """The FilePredicate filter function

        The arg slot expects a tuple of node_type and modpath.
        The predicate returns None (= agnostic about filtering) if
        the node is not a directory, otherwise it composites the
        modpath into a file path and applies it to the rest of
        the args
        """
        (node_type, modpath, module) = node_type__modpath__module
        if node_type == cellprofiler_core.setting.FileCollectionDisplay.NODE_DIRECTORY:
            return None
        elif isinstance(modpath[-1], tuple) and len(modpath[-1]) == 3:
            filename = modpath[-2]
        else:
            filename = modpath[-1]
        return args[0](filename, *args[1:])

    def test_valid(self, pipeline, *args):
        self(
            (
                cellprofiler_core.setting.FileCollectionDisplay.NODE_FILE,
                ["/imaging", "test.tif"],
                None,
            ),
            *args,
        )



from cellprofiler_core.constants.worker import ED_STOP, ED_SKIP

from ..pipeline import RunException


class PipelineEventListener:
    """listen for pipeline events, communicate them as necessary to the
    analysis manager."""

    def __init__(self, handle_exception_fn):
        self.handle_exception_fn = handle_exception_fn
        self.image_set_number = 0
        self.should_abort = False
        self.should_skip = False

    def reset(self):
        self.should_abort = False
        self.should_skip = False

    def handle_event(self, pipeline, event):
        if isinstance(event, RunException):
            disposition = self.handle_exception_fn(
                image_set_number=self.image_set_number,
                module_name=event.module.module_name,
                exc_info=(type(event.error), event.error, event.tb),
            )
            if disposition == ED_STOP:
                self.should_abort = True
                event.cancel_run = True
            elif disposition == ED_SKIP:
                self.should_skip = True
                event.cancel_run = False
                event.skip_thisset = True



from ._setting import Setting


class Binary(Setting):
    """A setting that is represented as either true or false
    The underlying value stored in the settings slot is "Yes" or "No"
    for historical reasons.
    """

    def __init__(self, text, value, callback=None, *args, **kwargs):
        """Initialize the binary setting with the module, explanatory
        text and value. The value for a binary setting is True or
        False.
        """
        str_value = (value and "Yes") or "No"
        super(Binary, self).__init__(text, str_value, *args, **kwargs)
        self.callback = callback

    def set_value(self, value):
        """When setting, translate true and false into yes and no"""
        if value in ("Yes", "No") or isinstance(value, str):
            super(Binary, self).set_value(value)
        else:
            str_value = (value and "Yes") or "No"
            super(Binary, self).set_value(str_value)

    def get_value(self):
        """Get the value of a binary setting as a truth value
        """
        return super(Binary, self).get_value() == "Yes"

    def eq(self, x):
        if x == "No":
            x = False
        return (self.value and x) or (not self.value and not x)

    def __bool__(self):
        """Return the value when testing for True / False"""
        return self.value

    def on_event_fired(self, selection):
        if self.callback is not None:
            self.callback(selection)



from ._file_image import FileImage


class STKFrameImage(FileImage):
    """Provide an image by filename:frame from an STK file"""

    def __init__(self, name, pathname, filename, frame, rescale_range=None):
        """Initialize the provider

        name - name of the provider for access from image set
        pathname - path to the file
        filename - name of the file
        frame - # of the frame to provide
        """
        super(STKFrameImage, self).__init__(
            name,
            pathname,
            filename,
            rescale_range=rescale_range,
            metadata_rescale=False,
            index=frame
        )



class ValidationError(ValueError):
    """An exception indicating that a setting's value prevents the pipeline from running
    """

    def __init__(self, message, setting):
        """Initialize with an explanatory message and the setting that caused the problem
        """
        super(ValidationError, self).__init__(message)
        self.message = message
        self.__setting = setting

    def get_setting(self):
        """The setting responsible for the problem

        This might be one of several settings partially responsible
        for the problem.
        """
        return self.__setting

    setting = property(get_setting)



from ..module import Module
from ..setting.text import Text


class MeasurementFixture(Module):
    module_name = "MeasurementFixture"
    category = "Test"
    variable_revision_number = 1
    do_not_check = True

    def create_settings(self):
        self.my_variable = Text("", "")
        self.other_variable = Text("", "")

    def settings(self):
        return [self.my_variable, self.other_variable]

    def module_class(self):
        return "cellprofiler_core.modules.measurementfixture.MeasurementFixture"

    def get_measurement_columns(self, pipeline):
        return [("Image", self.my_variable.value, "varchar(255)",)]



from ._crop_image_subscriber import CropImageSubscriber
from ._file_image_subscriber import FileImageSubscriber
from ._image_subscriber import ImageSubscriber
from ._outline_image_subscriber import OutlineImageSubscriber



import os
import urllib.parse
import urllib.request
import logging
from functools import cached_property

import numpy

from cellprofiler_core.constants.image import MD_SIZE_S, MD_SIZE_C, MD_SIZE_Z, MD_SIZE_T, MD_SIZE_Y, MD_SIZE_X, \
    MD_SIZE_KEYS, MD_SERIES_NAME
from cellprofiler_core.constants.modules.metadata import COL_PATH, COL_SERIES, COL_INDEX, COL_URL
from cellprofiler_core.constants.measurement import RESERVED_METADATA_KEYS
from cellprofiler_core.reader import get_image_reader, Reader
from cellprofiler_core.utilities.image import url_to_modpath, is_file_url

LOGGER = logging.getLogger(__name__)


class ImageFile:
    """This class represents an image file

    A file is considered as a single URL.

    A single file can contain multiple planes.

    This class assists in extracting information about the contents of
    a single file.

    """

    def __init__(self, url):
        self._url = url
        self._extracted = False
        self._index_mode = False
        self._reader = None
        self._metadata_dict = {
            COL_PATH: self.path,
            COL_SERIES: 0,
            COL_INDEX: 0,
            COL_URL: url,
            MD_SIZE_S: 0,
            MD_SIZE_C: [],
            MD_SIZE_Z: [],
            MD_SIZE_T: [],
            MD_SIZE_Y: [],
            MD_SIZE_X: [],
            MD_SERIES_NAME: [],
        }
        self._plane_details = []
        self._modpath = None
        # Records the name of the reader class selected for this file.
        self.preferred_reader = None

    def __repr__(self):
        return f"ImageFile object for {self.url}. Metadata extracted:{self.extracted}, Indexed:{self.index_mode}"

    def __lt__(self, other):
        if isinstance(other, str):
            return self.url < other
        elif isinstance(other, ImageFile):
            return self.url < other.url
        else:
            raise NotImplementedError("Unsupported comparison")

    def __gt__(self, other):
        if isinstance(other, str):
            return self.url > other
        elif isinstance(other, ImageFile):
            return self.url > other.url
        else:
            raise NotImplementedError("Unsupported comparison")

    def __eq__(self, other):
        if isinstance(other, str):
            return self.url == other
        elif isinstance(other, ImageFile):
            return self.url == other.url
        else:
            raise NotImplementedError("Unsupported comparison")

    def extract_planes(self, workspace=None):
        if self._extracted:
            return
        # Figure out the number of planes, indexes or series in the file.
        try:
            reader = self.get_reader()
        except:
            LOGGER.error(f"May not be an image: {self.url}", exc_info=True)
            self.metadata[MD_SIZE_S] = 0
            return
        meta_dict = reader.get_series_metadata()
        if meta_dict[MD_SIZE_S] == 0:
            LOGGER.error(f"File {self.filename} appears to contain no images.")
            self.metadata[MD_SIZE_S] = 0
            self.release_reader()
            return
        elif MD_SERIES_NAME not in meta_dict:
            meta_dict[MD_SERIES_NAME] = [''] * meta_dict[MD_SIZE_S]
        assert MD_SIZE_KEYS.issubset(meta_dict.keys()), "Returned metadata keys are incomplete"
        self.metadata.update(meta_dict)
        for S, C, Z, T, Y, X, name in self.get_plane_iterator():
            self._plane_details.append(f"Series {S:>2}{f' ({name})' if name else ''}"
                                       f": {X:>5} x {Y:<5}, {C} Channels, {Z:>2} Planes, {T:>2} Timepoints")
        if workspace is not None:
            metadata_array = numpy.transpose([
                self.metadata[MD_SIZE_C],
                self.metadata[MD_SIZE_Z],
                self.metadata[MD_SIZE_T],
                self.metadata[MD_SIZE_Y],
                self.metadata[MD_SIZE_X]])
            names_array = self.metadata[MD_SERIES_NAME]
            workspace.file_list.add_metadata(self.url, metadata_array.flatten(), names_array)
        self._extracted = True
        self.release_reader()

    def load_plane_metadata(self, data, names=''):
        # Metadata is stored in the HDF5 file as an array of int values, 5 per series for axis sizes (CZTYX)
        if len(data) < 5 or len(data) % 5 != 0:
            # No metadata or bad format
            if len(data) > 0:
                LOGGER.warning(f"Unable to load saved metadata for {self.filename}")
            return
        if len(data) == 5 and numpy.all(data == -1):
            # Unfilled metadata
            return
        num_series = len(data) // 5
        self.metadata[MD_SIZE_S] = num_series
        self.metadata[MD_SERIES_NAME] = names
        for i in range(num_series):
            C, Z, T, Y, X = data[:5]
            self.metadata[MD_SIZE_C].append(int(C))
            self.metadata[MD_SIZE_Z].append(int(Z))
            self.metadata[MD_SIZE_T].append(int(T))
            self.metadata[MD_SIZE_Y].append(int(Y))
            self.metadata[MD_SIZE_X].append(int(X))
        for S, C, Z, T, Y, X, name in self.get_plane_iterator():
            self._plane_details.append(
                f"Series {S:>2}{f' ({name})' if name else ''}"
                f": {X:>5} x {Y:<5}, {C} Channels, {Z:>2} Planes, {T:>2} Timepoints")
        self._extracted = True

    @property
    def extracted(self):
        return self._extracted

    @property
    def index_mode(self):
        return self._index_mode

    @property
    def url(self):
        return self._url

    @property
    def scheme(self):
        protocol_idx = self._url.find(":")
        if protocol_idx >= 0:
            return self._url.lower()[0: protocol_idx]
        else:
            return None

    @cached_property
    def filename(self):
        return os.path.basename(self.path)

    @cached_property
    def dirname(self):
        return os.path.dirname(self.path)

    @cached_property
    def path(self):
        """The file path if a file: URL, otherwise the URL"""
        if is_file_url(self.url):
            parsed = urllib.parse.urlparse(self.url)
            return urllib.request.url2pathname(parsed.path)
        return self.url

    @cached_property
    def modpath(self):
        """The directory, filename and extension broken up into a tuple"""
        if self._modpath is None:
            self._modpath = url_to_modpath(self.url)
        return self._modpath

    @cached_property
    def file_extension(self):
        return os.path.splitext(self.path)[-1].lower()

    @cached_property
    def full_extension(self):
        file_name = os.path.basename(self.path).lower()
        return file_name[file_name.find('.'):]

    @property
    def metadata(self):
        return self._metadata_dict

    def get_reader(self):
        if self._reader is None:
            self._reader = get_image_reader(self)
        return self._reader

    def release_reader(self):
        if self._reader is not None:
            if isinstance(self._reader, Reader):
                self._reader.close()
            self._reader = None

    def get_plane_iterator(self):
        # Returns an iterator which provides an entry for each series
        # in the file consisting of a tuple of the Series number, C, Z, T, Y and X dimension sizes.
        return zip(range(self.metadata[MD_SIZE_S]), self.metadata[MD_SIZE_C], self.metadata[MD_SIZE_Z],
                   self.metadata[MD_SIZE_T], self.metadata[MD_SIZE_Y], self.metadata[MD_SIZE_X],
                   self.metadata[MD_SERIES_NAME])

    @property
    def plane_details_text(self):
        return self._plane_details

    def add_metadata(self, meta_dict):
        # Used to bulk-add metadata keys from the Metadata module. Other modules should use set_metadata method.
        for key, value in meta_dict.items():
            if key in RESERVED_METADATA_KEYS:
                LOGGER.error(f"Unable to set protected metadata key '{key}' to '{value}' for file {self.filename}. "
                             f"Please choose another key name.")
                continue
            self._metadata_dict[key] = value

    def set_metadata(self, key, value, force=False):
        if key in RESERVED_METADATA_KEYS and not force:
            raise PermissionError(f"Cannot override protected metadata key '{key}'")
        else:
            if force:
                LOGGER.warning(f"Overwriting protected key {key}. This may break functionality.")
            self._metadata_dict[key] = value

    def clear_metadata(self):
        keep = (COL_PATH, COL_URL, COL_SERIES, COL_INDEX, MD_SIZE_S,
                MD_SIZE_C, MD_SIZE_Z, MD_SIZE_T, MD_SIZE_Y, MD_SIZE_X, MD_SERIES_NAME)
        self._metadata_dict = {k: self._metadata_dict[k] for k in keep}



import json
import logging
from packaging.version import Version

from cellprofiler_core import __version__ as core_version
from cellprofiler_core.constants.pipeline import (
    IMAGE_PLANE_DESCRIPTOR_VERSION,
    H_PLANE_COUNT,
)

LOGGER = logging.getLogger(__name__)

def dump(pipeline, fp, save_image_plane_details):
    """Serializes pipeline into JSON"""
    modules = []

    for module in pipeline.modules(False):
        settings = []
        for setting in module.settings():
            settings.append(setting.to_dict())

        modules += [{"attributes": module.to_dict(), "settings": settings}]

    if len(pipeline.file_list) == 0:
        save_image_plane_details = False
        
    sem_ver = Version(core_version)

    content = {
        "has_image_plane_details": save_image_plane_details,
        "date_revision": int(f"{sem_ver.major}{sem_ver.minor}{sem_ver.micro}"),
        "module_count": len(pipeline.modules(False)),
        "modules": modules,
        "version": "v6",
    }

    if save_image_plane_details:
        urls = [url for url in pipeline.file_list]
        file_list = {
            "version": '"%s":"%d","%s":"%d"'
            % (
                "Version",
                IMAGE_PLANE_DESCRIPTOR_VERSION,
                H_PLANE_COUNT,
                len(pipeline.file_list),
            ),
            "urls": urls,
        }
        content["file_list"] = file_list

    json.dump(content, fp, indent=4)


def load(pipeline, fd):
    pipeline_dict = json.load(fd)
    sem_ver = Version(core_version)
    cp_version = int(f"{sem_ver.major}{sem_ver.minor}{sem_ver.micro}")
    if cp_version != pipeline_dict['date_revision']:
        LOGGER.warning(f"Pipeline file is from a different version of CellProfiler. "
                        f"Current:v{cp_version} File:v{pipeline_dict['date_revision']}."
                        f" Will attempt to upgrade settings.")
    pipeline_modules = pipeline.modules(False)
    pipeline_modules.clear()
    for module in pipeline_dict["modules"]:
        module_name = module["attributes"]["module_name"]
        settings = [setting_dict for setting_dict in module["settings"]]
        new_module = pipeline.instantiate_module(module_name)
        new_module.from_dict(settings, module["attributes"])
        pipeline_modules.append(new_module)




from ._event import Event


class IPDLoadException(Event):
    """An exception was cauaght while trying to load the image plane details

    This event is reported when an exception is thrown while loading
    the image plane details from the workspace's file list.
    """

    def __init__(self, error):
        super(self.__class__, self).__init__()
        self.error = error
        self.cancel_run = True

    def event_type(self):
        return "Image load exception"



import numpy


class GrayscaleImage:
    """A wrapper around a non-grayscale image

    This is meant to be used if the image is 3-d but all channels
       are the same or if the image is binary.
    """

    def __init__(self, image):
        self.__image = image

    def __getattr__(self, name):
        return getattr(self.__image, name)

    @property
    def pixel_data(self):
        """One 2-d channel of the color image as a numpy array"""
        if self.__image.pixel_data.dtype.kind == "b":
            return self.__image.pixel_data.astype(numpy.float64)

        return self.__image.pixel_data[:, :, 0]
